﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

using BenLincoln.DebugSymbols.SymDump;

namespace BenLincoln.DebugSymbols
{
    public struct AddDataTypeByNameResult
    {
        public string GhidraCodeSubroutines;
        public string GhidraCodeSubroutineCalls;
        public List<string> AddedNames;
        public int NextSubroutineNumber;
    }

    public struct RenameJSONEntriesForGhidraResult
    {
        public string JSONText;
        public StringBuilder ManualChangesBuilder;
    }

    public class GhidraScriptGenerator
    {

        public static AddDataTypeByNameResult GetAddDataTypesScriptCode(string ghidraImportDataTypeSubroutineTemplate, IEnumerable<BenLincoln.DebugSymbols.SymDump.CElement> elements, string projectName, string elementTypeName, int maxElementsPerSubroutine, int currentSubroutineNumber)
        {
            StringBuilder subCollectionBuilder = new StringBuilder();
            StringBuilder subBuilder = new StringBuilder();
            StringBuilder subCallBuilder = new StringBuilder();
            AddDataTypeByNameResult result = new AddDataTypeByNameResult();
            result.NextSubroutineNumber = currentSubroutineNumber;
            int dtNum = 0;
            int elemNum = 0;
            result.AddedNames = new List<string>();
            foreach (CElement e in elements)
            {
                if (!result.AddedNames.Contains(e.name))
                {
                    subBuilder.Append(String.Format("\t\tdataTypeMap = addDataTypeByName(dataTypeMap, \"/{0}.H/{1}\", \"{1}\", \"{2}\");", projectName, BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(e.name), BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeJavaString(elementTypeName)));
                    subBuilder.Append(Environment.NewLine);
                    result.AddedNames.Add(e.name);
                    dtNum++;
                }
                elemNum++;
                if ((dtNum > maxElementsPerSubroutine) || (elemNum >= elements.Count()))
                {
                    string subRoutineDefinition = ghidraImportDataTypeSubroutineTemplate.Replace("%IMPORT_DATATYPES_SUB_CONTENT%", subBuilder.ToString());
                    subRoutineDefinition = subRoutineDefinition.Replace("%IMPORT_DATATYPES_SUBROUTINE_NUMBER%", String.Format("{0}", result.NextSubroutineNumber));
                    subCollectionBuilder.Append(subRoutineDefinition);
                    subCallBuilder.Append(String.Format("\t\tdataTypeMap = ImportDataTypesSub{0}(dataTypeMap);", result.NextSubroutineNumber));
                    subCallBuilder.Append(Environment.NewLine);
                    subBuilder.Clear();
                    dtNum = 0;
                    result.NextSubroutineNumber++;
                }                
            }
            result.GhidraCodeSubroutines = subCollectionBuilder.ToString();
            result.GhidraCodeSubroutineCalls = subCallBuilder.ToString();
            
            return result;
        }

        public static RenameJSONEntriesForGhidraResult RenameJSONEntriesForGhidra(string jsonText, StringBuilder manualChangesBuilder)
        {
            // fix names that don't work in Ghidra
            RenameJSONEntriesForGhidraResult result = new RenameJSONEntriesForGhidraResult();
            result.JSONText = jsonText;

            Regex dotRenameRex = new Regex("(?<itemName>(union|struct|enum)( | [^ ]+ \\*)\\.[^ ;\"$]+)[ ;\"$]");
            MatchCollection drMatches = dotRenameRex.Matches(result.JSONText);
            Dictionary<string, int> dotNames = new Dictionary<string, int>();
            foreach (Match m in drMatches)
            {
                if (m.Groups["itemName"] != null)
                {
                    string matchText = m.Groups["itemName"].Value;
                    if (!dotNames.ContainsKey(matchText))
                    {
                        dotNames.Add(matchText, 1);
                    }
                    else
                    {
                        dotNames[matchText]++;
                    }
                }
            }

            if (dotNames.Count > 0)
            {
                manualChangesBuilder.Append("The input JSON file contains at least one C element with a name that starts with a '.'. This naming convention is incompatible with Ghidra, so it will be replaced by a leading underscore. You should consider renaming these elements in the source JSON file, and will need to do so if any of them have name conflicts listed later in this file.");
                manualChangesBuilder.Append(Environment.NewLine);
                List<string> keyNames = dotNames.Keys.ToList();
                keyNames.Sort();
                manualChangesBuilder.Append("Elements which will be renamed:");
                manualChangesBuilder.Append(Environment.NewLine);
                foreach (string dotName in keyNames)
                {
                    if (dotNames[dotName] > 1)
                    {
                        manualChangesBuilder.Append(string.Format("\t{0} ({1} occurrences)", dotName, dotNames[dotName]));
                    }
                    else
                    {
                        manualChangesBuilder.Append(string.Format("\t{0}", dotName));
                    }
                    manualChangesBuilder.Append(Environment.NewLine);
                }
                manualChangesBuilder.Append(Environment.NewLine);
            }

            result.JSONText = result.JSONText.Replace("union .", "union _");
            result.JSONText = result.JSONText.Replace("struct .", "struct _");
            result.JSONText = result.JSONText.Replace("enum .", "enum _");
            result.JSONText = result.JSONText.Replace("\"name\": \".", "\"name\": \"_");
            foreach (string matchText in dotNames.Keys)
            {
                string replaceText = matchText.Replace(".", "_");
                Console.WriteLine(string.Format("Debug: replacing '{0}' with '{1}', {2} occurrences in JSON file", matchText, replaceText, dotNames[matchText]));
                result.JSONText = result.JSONText.Replace(matchText, replaceText);
            }

            result.ManualChangesBuilder = manualChangesBuilder;
            return result;
        }
    }
}
