﻿// This file is part of BenLincoln.DebugSymbols.

// BenLincoln.DebugSymbols is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.DebugSymbols is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.DebugSymbols (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace BenLincoln.DebugSymbols.SymDump
{



    public class CElement : IComparable<CElement>
    {
        public string name { get; set; }
        public string source_file { get; set; }
        public long source_file_line_number { get; set; }
        public List<string> object_files { get; set; }
        public long offset_in_memory { get; set; }
        public List<CFunction> used_by_functions;
        public int hashcode { get; set; }
        public int sym_file_sequence_number { get; set; }
        public int overlay_id { get; set; }
        public Guid guid { get; set; }
        public bool fake { get; set; }

        public CElement()
        {
            used_by_functions = new List<CFunction> { };
            name = "";
            source_file = "";
            source_file_line_number = -1;
            object_files = new List<string>();
            offset_in_memory = -1;
            sym_file_sequence_number = -1;
            hashcode = -1;
            overlay_id = -1;
            guid = Guid.NewGuid();
            fake = false;
        }

        public int CompareTo(CElement other)
        {
            if (sym_file_sequence_number != other.sym_file_sequence_number)
            {
                return sym_file_sequence_number.CompareTo(other.sym_file_sequence_number);
            }
            if (name != other.name)
            {
                return name.CompareTo(other.name);
            }
            if (source_file != other.source_file)
            {
                return source_file.CompareTo(other.source_file);
            }
            if (source_file_line_number != other.source_file_line_number)
            {
                return source_file_line_number.CompareTo(other.source_file_line_number);
            }
            //if (object_file != other.object_file)
            //{
            //    return object_file.CompareTo(other.object_file);
            //}

            // fallback
            return name.CompareTo(other.name);
        }

        // replaces problematic characters in names
        public static string GetCompatibleName(string name, bool ignoreEOS)
        {
            if (ignoreEOS)
            {
                if (name == ".eos")
                {
                    return name;
                }
            }
            string result = name;
            result = result.Replace(".", "_");
            if (name != result)
            {
                //if (name == ".253fake")
                //{
                //    Console.WriteLine("Debug");
                //}
                Console.WriteLine(string.Format("Info: the item named '{0}' has been renamed '{1}' for better compatibility with most tools.", name, result));
            }
            return result;
        }

        public virtual string GetHeaderCode()
        {
            return "";
        }

        public virtual string GetSourceCode()
        {
            return "";
        }

        public static string GetGhidrafiedDataTypeName(string typeName)
        {
            string ghidraTypeName = typeName;
            if (ghidraTypeName.StartsWith("struct "))
            {
                ghidraTypeName = ghidraTypeName.Substring(7);
            }
            if (ghidraTypeName.StartsWith("union "))
            {
                ghidraTypeName = ghidraTypeName.Substring(6);
            }
            if (ghidraTypeName.StartsWith("enum "))
            {
                ghidraTypeName = ghidraTypeName.Substring(5);
            }
            return ghidraTypeName;
        }

        public static string GetHashCodeAsString(int hashcode)
        {
            return string.Format("0x{0:X8} (dec: {0})", hashcode);
        }

        public string GetHashCodeAsString()
        {
            return GetHashCodeAsString(hashcode);
        }

        public static string StripBlockCommentsFromString(string s)
        {
            //return s;
            Regex rex = new Regex(@"/\*[^\*]+\*/");
            //Regex rex = new Regex(@"/\*(?!\*/).+\*/");
            string result = rex.Replace(s, "");
            result = result.Replace(" ,", ",");
            while (result.Contains("  "))
            {
                result = result.Replace("  ", " ");
            }
            result = result.Replace(" )", ")");
            return result;
        }

        //public static CParameter[] GetCParametersFromString(string inputString, bool isParameterSection)
        //{
        //    // strip trailing comments
        //    string inputStringStripped = inputString;
        //    Regex trailingCommentRex = new Regex("//.+$");
        //    inputStringStripped = trailingCommentRex.Replace(inputStringStripped, "");

        //    // strip trailing semicolon
        //    Regex trailingSemicolonRex = new Regex(";.+$");
        //    inputStringStripped = trailingSemicolonRex.Replace(inputStringStripped, "");

        //    // hacky workaround for a Ghidra quirk of writing methods with no parameters as void SomeMethod(void) instead of void SomeMethod()
        //    if (inputStringStripped == "void)")
        //    {
        //        return new CParameter[0];
        //    }

        //    string[] paramStringArray = inputStringStripped.Split(new char[] { ',' });
        //    CParameter[] result = new CParameter[paramStringArray.Length];

        //    for (int paramNum = 0; paramNum < paramStringArray.Length; paramNum++)
        //    {
        //        result[paramNum] = new CParameter();
        //        result[paramNum].is_function_pointer = false;
        //        result[paramNum].is_varargs = false;
        //        result[paramNum].register_name = "";
        //        result[paramNum].name = "";

        //        string typeName = "";
        //        string[] paramArray = paramStringArray[paramNum].Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
        //        typeName = paramArray[0];
        //        int nextArrayIndex = 1;
        //        if ((typeName == "...") || (typeName == "...)"))
        //        {
        //            result[paramNum - 1].is_varargs = true;
        //            CParameter[] result2 = new CParameter[paramStringArray.Length - 1];
        //            for (int pCopyNum = 0; pCopyNum < result2.Length; pCopyNum++)
        //            {
        //                result2[pCopyNum] = result[pCopyNum];
        //                return result2;
        //            }
        //            break;
        //        }
        //        if (typeName == "struct" || typeName == "union" || typeName == "unsigned" || typeName == "enum")
        //        {
        //            typeName = String.Format("{0} {1}", typeName, paramArray[1]);
        //            nextArrayIndex = 2;
        //        }
        //        if (typeName == "unsigned long")
        //        {
        //            if (paramArray.Length > nextArrayIndex)
        //            {
        //                if (paramArray[nextArrayIndex] == "long")
        //                {
        //                    typeName = "unsigned long long";
        //                    nextArrayIndex++;
        //                }
        //            }
        //        }
        //        if (paramArray.Length > nextArrayIndex)
        //        {
        //            // handle both return type syntax (void * methodName)
        //            // and parameter syntax (void *paramName)
        //            // and function pointer syntax (void (*something()))
        //            bool handled = false;
        //            if (paramArray[nextArrayIndex].StartsWith("*"))
        //            {
        //                if (paramArray[nextArrayIndex].StartsWith("**"))
        //                {
        //                    typeName = String.Format("{0} **", typeName);
        //                    if (paramArray[nextArrayIndex] == "**")
        //                    {
        //                        nextArrayIndex++;
        //                    }
        //                    else
        //                    {
        //                        paramArray[nextArrayIndex] = paramArray[nextArrayIndex].Substring(2);
        //                    }
        //                }
        //                else
        //                {
        //                    typeName = String.Format("{0} *", typeName);
        //                    if (paramArray[nextArrayIndex] == "*")
        //                    {
        //                        nextArrayIndex++;
        //                    }
        //                    else
        //                    {
        //                        paramArray[nextArrayIndex] = paramArray[nextArrayIndex].Substring(1);
        //                    }
        //                }
        //                handled = true;
        //            }
        //            if (!handled)
        //            {
        //                if (paramArray[nextArrayIndex].StartsWith("(*"))
        //                {
        //                    if (paramArray[nextArrayIndex].EndsWith(")"))
        //                    {
        //                            result[paramNum].is_function_pointer = true;
        //                            //string functionPointerName = paramArray[nextArrayIndex];
        //                            result[paramNum].name = paramArray[nextArrayIndex];
        //                            typeName = String.Format("{0} *", typeName);
        //                            //if (functionPointerName.Length > 5)
        //                            //{
        //                            //    functionPointerName = functionPointerName.Substring(2);
        //                            //    functionPointerName = functionPointerName.Substring(0, functionPointerName.Length - 3);
        //                            //    functionPointerName = String.Format("{0} /* <= function pointer */", functionPointerName);
        //                            //    result[paramNum].ParameterName = functionPointerName;
        //                            //}
        //                            nextArrayIndex++;
        //                            handled = true;
        //                    }
        //                    else
        //                    {
        //                        if (paramArray[nextArrayIndex].Contains("["))
        //                        {
        //                        }
        //                    }
        //                }
        //            }
        //        }
        //        result[paramNum].type_name = typeName;
        //        if (paramArray.Length > nextArrayIndex)
        //        {
        //            if (paramArray[nextArrayIndex].StartsWith("/*") && paramArray[nextArrayIndex].EndsWith("*/"))
        //            {
        //                result[paramNum].register_name = paramArray[nextArrayIndex];
        //                result[paramNum].register_name = result[paramNum].register_name.Substring(2, result[paramNum].register_name.Length - 4);
        //                nextArrayIndex++;
        //            }
        //        }

        //        if ((result[paramNum].name == "") || (result[paramNum].name == "(*)()"))
        //        {
        //            if (nextArrayIndex >= paramArray.Length)
        //            {
        //                throw new Exception(String.Format("Parsing error: nextArrayIndex ({0}) is past the end of the array in the string '{1}', but no method name has been parsed.", nextArrayIndex, paramStringArray[paramNum]));
        //            }

        //            if (paramArray.Length > nextArrayIndex)
        //            {
        //                result[paramNum].name = paramArray[nextArrayIndex];
        //                nextArrayIndex++;
        //            }
        //        }

        //        if (paramArray.Length > nextArrayIndex)
        //        {
        //            if (paramArray[nextArrayIndex].StartsWith("/*") && paramArray[nextArrayIndex].EndsWith("*/"))
        //            {
        //                result[paramNum].register_name = paramArray[nextArrayIndex];
        //                result[paramNum].register_name = result[paramNum].register_name.Substring(2, result[paramNum].register_name.Length - 4);
        //                nextArrayIndex++;
        //            }
        //        }
        //        //if (result[paramNum].ParameterName.StartsWith("(*"))
        //        //{
        //        //    if (result[paramNum].ParameterName.EndsWith(")()"))
        //        //    {
        //        //        result[paramNum].FunctionPointer = true;
        //        //    }
        //        //}
        //        if (result[paramNum].name.StartsWith("(*"))
        //        {                    
        //            if (result[paramNum].name.EndsWith(")"))
        //            {
        //                result[paramNum].is_function_pointer = true;
        //                string originalParamName = result[paramNum].name;
        //                result[paramNum].name = result[paramNum].name.Substring(2);
        //                int firstClosingParenPosition = result[paramNum].name.IndexOf(')');
        //                result[paramNum].name = result[paramNum].name.Substring(0, firstClosingParenPosition);
        //                //result[paramNum].ParameterName = String.Format("{0} /* <= function pointer: {1} */", result[paramNum].ParameterName, originalParamName);
        //            }
        //            //else
        //            //{
        //            //    string msg = String.Format("Found a string that seems to be a function pointer ('{0}'), but does not end in ')'.", result[paramNum].name);
        //            //    throw new Exception(msg);
        //            //}
        //        }

        //        // catch arrays
        //        if (result[paramNum].name.Contains("["))
        //        {
        //            if (result[paramNum].name.Contains("]"))
        //            {
        //                int bracketOpenPos = result[paramNum].name.IndexOf("[");
        //                string arraySize = result[paramNum].name.Substring(bracketOpenPos);
        //                result[paramNum].name = result[paramNum].name.Substring(0, bracketOpenPos);
        //                result[paramNum].type_name = String.Format("{0}{1}", result[paramNum].type_name, arraySize);
        //            }
        //            else
        //            {
        //                throw new Exception(String.Format("Found a parameter name that contains '[', ('{0}'), but not ']'.", result[paramNum].name));
        //            }
        //        }

        //        // not sure why these aren't getting picked up properly yet
        //        //if (result[paramNum].type_name.EndsWith("void (*)()"))
        //        //{
        //        //    result[paramNum].is_function_pointer = true;
        //        //    // these need to be assigned a name because Ghidra can't really refer to them otherwise, at least not in any way I know of
        //        //    result[paramNum].type_name = result[paramNum].type_name.Substring(0, result[paramNum].type_name.Length - 6);
        //        //    if (result[paramNum].name == "")
        //        //    {
        //        //        result[paramNum].name = string.Format("unnamed_{0}_function_pointer", result[paramNum].type_name);
        //        //    }
        //        //}

        //        if (result[paramNum].name.EndsWith("("))
        //        {
        //            result[paramNum].name = result[paramNum].name.Substring(0, result[paramNum].name.Length - 1);
        //        }

        //    }

        //    return result;
        //}

        public static string GhidraXMLEncodeString(string InputString)
        {
            string result = InputString;

            result = result.Replace("&", "&amp;");
            result = result.Replace("'", "&apos;");
            result = result.Replace("<", "&lt;");
            result = result.Replace(">", "&gt;");
            result = result.Replace("\"", "&quot;");

            return result;
        }
    }

    public class CEnum : CElement
    {
        public string primitive_type { get; set; }
        public Dictionary<string, int> members { get; set; }

        public CEnum()
            : base()
        {
            primitive_type = "";
            members = new Dictionary<string, int>();
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public override string GetHeaderCode()
        {
            StringBuilder builder = new StringBuilder();
            builder.Append(String.Format("enum {0} // Hashcode: {1}", BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeCNameWithDefaultReplacement(name), GetHashCodeAsString()));
            builder.Append(Environment.NewLine);
            builder.Append("{");
            builder.Append(Environment.NewLine);
            foreach (string valueName in members.Keys)
            {
                builder.Append(String.Format("\t{0} = {1}", BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeCNameWithDefaultReplacement(valueName), members[valueName]));
                builder.Append(",");
                builder.Append(Environment.NewLine);
            }
            builder.Append("};");
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            string lengthString = "0x4";
            if (primitive_type != "int")
            {
                throw new Exception(String.Format("Unsupported enum primitive type: {0}", primitive_type));
            }

            builder.Append(indent);
            builder.Append(String.Format("<enum name=\"{0}\" type=\"{1}\" length=\"{2}\">", GhidraXMLEncodeString(name), primitive_type, lengthString));
            builder.Append(Environment.NewLine);

            foreach (string memberName in members.Keys)
            {
                builder.Append(indent);
                builder.Append(String.Format("\t<member name=\"{0}\" value=\"0x{1:X}\" />", GhidraXMLEncodeString(memberName), members[memberName]));
                builder.Append(Environment.NewLine);
            }

            builder.Append(indent);
            builder.Append("</enum>");
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class CStructMember : CElement
    {
        public string struct_member_signature { get; set; }
        public string class_type { get; set; }
        public string c_type { get; set; }
        public string type_name { get; set; }
        public uint size { get; set; }
        //public long offset { get; set; }
        public string parent_name { get; set; }
        public int parent_hashcode { get; set; }
        public string[] derived_types { get; set; }

        public CStructMember()
            : base()
        {
            struct_member_signature = "";
            class_type = "";
            c_type = "";
            type_name = "";
            size = 0;
            //offset = -1;
            parent_name = "";
            parent_hashcode = -1;
            derived_types = new string[] { };
        }
    }

    public class CStruct : CElement
    {
        public List<CStructMember> members { get; set; }

        public CStruct()
            : base()
        {
            members = new List<CStructMember>();
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public uint GetSize()
        {
            uint result = 0;
            for (int i = 0; i < members.Count; i++)
            {
                result += members[i].size;
            }
            return result;
        }

        public override string GetHeaderCode()
        {
            StringBuilder builder = new StringBuilder();
            builder.Append(String.Format("struct {0} // hashcode: {1}", BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeCNameWithDefaultReplacement(name), GetHashCodeAsString()));
            builder.Append(Environment.NewLine);
            builder.Append("{");
            builder.Append(Environment.NewLine);
            foreach (CStructMember mem in members)
            {
                //builder.Append("\t");
                //builder.Append(mem.struct_member_signature);
                //builder.Append(string.Format(" //  hashcode: {0}, parent name: '{1}', parent hashcode: {2}", mem.GetHashCodeAsString(), mem.parent_name, CElement.GetHashCodeAsString(mem.parent_hashcode)));
                //builder.Append(Environment.NewLine);
                string[] lines = mem.struct_member_signature.Split(new string[] { Environment.NewLine, "\r", "\n" }, StringSplitOptions.RemoveEmptyEntries);
                int numTabs = 1;
                for (int lNum = 0; lNum < lines.Length; lNum++)
                {
                    if (lines[lNum].Trim().StartsWith("}"))
                    {
                        numTabs--;
                    }
                    for (int tNum = 0; tNum < numTabs; tNum++)
                    {
                        builder.Append("\t");
                    }
                    builder.Append(lines[lNum].Trim());
                    //builder.Append(mem.struct_member_signature);
                    builder.Append(Environment.NewLine);
                    if (lines[lNum].Trim().EndsWith("{"))
                    {
                        numTabs++;
                    }
                }
            }
            builder.Append("};");
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            uint totalLength = 0;
            foreach (CStructMember m in members)
            {
                totalLength += m.size;
            }

            builder.Append(indent);
            builder.Append(String.Format("<datatype name=\"{0}\" kind=\"Structure\" length=\"0x{1:X}\" >", GhidraXMLEncodeString(name), totalLength));
            builder.Append(Environment.NewLine);

            for (int i = 0; i < members.Count; i++)
            {
                builder.Append(indent);
                string ghidraTypeName = GetGhidrafiedDataTypeName(members[i].type_name);

                builder.Append(String.Format("\t<member name=\"{0}\" datatype=\"{1}\" offset=\"0x{2:X8}\" kind=\"Member\" length=\"0x{3:X}\" />", 
                    GhidraXMLEncodeString(members[i].name), GhidraXMLEncodeString(ghidraTypeName), members[i].offset_in_memory, members[i].size));
                builder.Append(Environment.NewLine);
            }

            builder.Append(indent);
            builder.Append("</datatype>");
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class CUnion : CElement
    {
        public List<CStructMember> members { get; set; }

        public CUnion()
            : base()
        {
            members = new List<CStructMember>();
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public uint GetSize()
        {
            uint result = 0;
            for (int i = 0; i < members.Count; i++)
            {
                result += members[i].size;
            }
            return result;
        }

        public override string GetHeaderCode()
        {
            StringBuilder builder = new StringBuilder();
            builder.Append(String.Format("union {0} // Hashcode: {1}", BenLincoln.DebugSymbols.InputValidator.GetReasonablySafeCNameWithDefaultReplacement(name), GetHashCodeAsString()));
            builder.Append(Environment.NewLine);
            builder.Append("{");
            builder.Append(Environment.NewLine);
            foreach (CStructMember mem in members)
            {
                string[] lines = mem.struct_member_signature.Split(new string[] { Environment.NewLine, "\r", "\n" }, StringSplitOptions.RemoveEmptyEntries);
                int numTabs = 1;
                for (int lNum = 0; lNum < lines.Length; lNum++)
                {
                    //if (lNum > 1)
                    //{
                    //    Console.WriteLine("Debug");
                    //}
                    if (lines[lNum].Trim().StartsWith("}"))
                    {
                        numTabs--;
                    }
                    for (int tNum = 0; tNum < numTabs; tNum++)
                    {
                        builder.Append("\t");
                    }
                    builder.Append(lines[lNum].Trim());
                    //builder.Append(mem.struct_member_signature);
                    builder.Append(Environment.NewLine);
                    if (lines[lNum].Trim().EndsWith("{"))
                    {
                        numTabs++;
                    }
                }
            }
            builder.Append("};");
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            uint totalLength = 0;
            // unions are an overlaid memory structure whose size is the greatest of any individual member
            foreach (CStructMember m in members)
            {
                if (m.size > totalLength)
                {
                    totalLength = m.size;
                }
            }

            builder.Append(indent);
            builder.Append(String.Format("<datatype name=\"{0}\" kind=\"Union\" length=\"0x{1:X}\" >", GhidraXMLEncodeString(name), totalLength));
            builder.Append(Environment.NewLine);

            for (int i = 0; i < members.Count; i++)
            {
                builder.Append(indent);
                string ghidraTypeName = GetGhidrafiedDataTypeName(members[i].type_name);

                builder.Append(String.Format("\t<member name=\"{0}\" datatype=\"{1}\" offset=\"0x{2:X8}\" kind=\"Member\" length=\"0x{3:X}\" />", 
                    GhidraXMLEncodeString(members[i].name), GhidraXMLEncodeString(ghidraTypeName), members[i].offset_in_memory, members[i].size));
                builder.Append(Environment.NewLine);
            }

            builder.Append(indent);
            builder.Append("</datatype>");
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class CTypeDef : CElement
    {
        public string c_type { get; set; }
        public string typedef_signature { get; set; }

        public CTypeDef()
            : base()
        {
            c_type = "";
            typedef_signature = "";
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public override string GetHeaderCode()
        {
            StringBuilder builder = new StringBuilder();
            builder.Append(String.Format("typedef {0};", typedef_signature));
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(indent);
            builder.Append(String.Format("<typedef name=\"{0}\" basetype=\"{1}\" />", name, c_type));
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class CLabel : CElement
    {
        //public long offset { get; set; }
        public int label_type { get; set; }

        public CLabel()
            : base()
        {
            //offset = -1;
            label_type = -1;
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public string ToGhidraSymbolXML(string indent, int symbolIndex)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(indent);
            builder.Append(String.Format("<symbol name =\"{0}\" address=\"0x{1:X}\" length=\"0x0\" tag=\"Label\" kind=\"Unknown\" index=\"0x{2:X}\" undecorated=\"\" value=\"\" datatype=\"\"  />", 
                name, offset_in_memory, symbolIndex));
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class CExternalDeclaration : CElement
    {
        public string declaration { get; set; }
        public string extern_type { get; set; }
        public string class_type { get; set; }
        public long size { get; set; }
        //public long offset { get; set; }
        public string source_code { get; set; }

        public CExternalDeclaration()
            : base()
        {
            declaration = "";
            extern_type = "";
            class_type = "";
            size = -1;
            //offset = -1;
            source_code = "";
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public override string GetHeaderCode()
        {
            return declaration + Environment.NewLine;
        }

        public override string GetSourceCode()
        {
            //if ((source_code == null) || (source_code == ""))
            //{
            //    return string.Format("// {0} (@0x{1:X8}, '{2}') is not initialized, or was not found by TDR, or was not identified in Ghidra", name, offset, declaration);
            //}
            return source_code;
        }

        public string GetName()
        {
            string result = declaration;
            if (result.Contains(";"))
            {
                int semiColonPos = result.IndexOf(";");
                if (semiColonPos > -1)
                {
                    result = result.Substring(0, semiColonPos);
                }
            }
            int lastSpacePos = result.LastIndexOf(" ");
            if (lastSpacePos > -1)
            {
                result = result.Substring(lastSpacePos + 1);
            }
            // trim off any pointer *s
            while (result.StartsWith("*"))
            {
                result = result.Substring(1);
            }
            // trim off any size definition
            if (result.EndsWith("]"))
            {
                int openBracketPos = result.IndexOf("[");
                if (openBracketPos > -1)
                {
                    result = result.Substring(0, openBracketPos);
                }
            }
            return result;
        }

        public string GetTypeName()
        {
            string result = declaration;
            if (result.Contains(";"))
            {
                int semiColonPos = result.IndexOf(";");
                if (semiColonPos > -1)
                {
                    result = result.Substring(0, semiColonPos);
                }
            }
            int lastSpacePos = result.LastIndexOf(" ");
            if (lastSpacePos > -1)
            {
                if (result.Contains("*"))
                {
                    lastSpacePos = result.LastIndexOf("*") + 1;
                }
                result = result.Substring(0, lastSpacePos);
            }
            // trim off "static ", "extern ", etc.
            if (result.StartsWith("static ") || result.StartsWith("extern "))
            {
                int firstSpacePos = result.IndexOf(" ");
                if (firstSpacePos > -1)
                {
                    result = result.Substring(firstSpacePos + 1);
                }
            }
            // fix unbalanced parens for pointers
            // this is a hack, sorry
            if (result.Contains("(*"))
            {
                if (result.EndsWith("*"))
                {
                    if (!result.Contains(")"))
                    {
                        result += ")";
                    }
                }
            }
            return result;
        }

        public int GetCount()
        {
            int result = 1;
            if (declaration.Contains("[") && declaration.Contains("]"))
            {
                string d = declaration;
                if (d.Contains(";"))
                {
                    int semiColonPos = d.IndexOf(";");
                    if (semiColonPos > -1)
                    {
                        d = d.Substring(0, semiColonPos);
                    }
                }
                int lastSpacePos = d.LastIndexOf(" ");
                if (lastSpacePos > -1)
                {
                    d = d.Substring(lastSpacePos + 1);
                }
                int openBracketPos = d.IndexOf("[");
                if (openBracketPos > -1)
                {
                    d = d.Substring(openBracketPos + 1);
                }
                int closeBracketPos = d.IndexOf("]");
                if (openBracketPos > -1)
                {
                    d = d.Substring(0, closeBracketPos);
                }
                try
                {
                    result = int.Parse(d);
                }
                catch (Exception e)
                {
                    throw new Exception(string.Format("Could not parse an array length from the declaration '{0}'", declaration));
                }
            }
            return result;
        }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            //builder.Append(indent);
            //builder.Append(String.Format("<datatype name=\"{0}\" kind=\"Structure\" length=\"0x10\" >", name, primitive_type, lengthString));
            //builder.Append(Environment.NewLine);

            //builder.Append(indent);
            //builder.Append("</datatype>");
            //builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class CVariable : CElement
    {
        // stack, register, or static
        public string variable_signature { get; set; }
        public string variable_type { get; set; }
        public string register_name { get; set; }
        public long stack_offset { get; set; }
        public long static_offset { get; set; }
        public string class_type { get; set; }
        public string c_type { get; set; }
        public string type_name { get; set; }
        public uint size { get; set; }

        public CVariable()
            : base()
        {
            variable_signature = "";
            variable_type = "";
            register_name = "";
            //stack_offset = -1;
            //static_offset = -1;
            class_type = "";
            c_type = "";
            type_name = "";
            size = 0;
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        //case ClassType.AutoVar:
        //    _vars.Add($"{ti.AsCode(memberName)}; // stack offset {typedValue.Value}");
        //    break;
        //case ClassType.Register:
        //    _vars.Add($"{ti.AsCode(memberName)}; // ${(Register) typedValue.Value}");
        //    break;
        //case ClassType.Static:
        //    _vars.AdAdd($"static {ti.AsCode(memberName)}; // offset 0x{typedValue.Value:x}");

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            //builder.Append(indent);
            //builder.Append(String.Format("<datatype name=\"{0}\" kind=\"Structure\" length=\"0x10\" >", name, primitive_type, lengthString));
            //builder.Append(Environment.NewLine);

            //builder.Append(indent);
            //builder.Append("</datatype>");
            //builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class CBlock : CElement
    {
        public long start_line { get; set; }
        public long end_line { get; set; }
        public long start_offset { get; set; }
        public long end_offset { get; set; }
        public List<CTypeDef> typedefs { get; set; }
        //public List<string> typedefs { get; set; }
        public List<CVariable> variables { get; set; }
        public List<CLabel> labels { get; set; }
        public List<CBlock> sub_blocks { get; set; }

        public CBlock()
            : base()
        {
            //start_line = -1;
            //end_line = -1;
            //start_offset = -1;
            //end_offset = -1;
            typedefs = new List<CTypeDef>();
            variables = new List<CVariable>();
            labels = new List<CLabel>();
            sub_blocks = new List<CBlock>();
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public string GetStubCode(CFunction parentMethod, string indent, string blockName, bool commentOutCode)
        {
            StringBuilder builder = new StringBuilder();
            builder.Append(indent);
            builder.Append(String.Format("/* begin block {0} */", blockName));
            builder.Append(Environment.NewLine);
            builder.Append(indent);
            builder.Append(String.Format("\t// Start line: {0}", parentMethod.source_file_line_number + start_line));
            builder.Append(Environment.NewLine);
            if (start_offset != 0)
            {
                builder.Append(indent);
                builder.Append(String.Format("\t// Start offset: 0x{0:X8}", start_offset));
                builder.Append(Environment.NewLine);
            }
            if (typedefs.Count > 0)
            {
                builder.Append(indent);
                builder.Append(String.Format("\t// TypeDefs:"));
                builder.Append(Environment.NewLine);
                foreach (CTypeDef t in typedefs)
                {
                    builder.Append(indent);
                    if (commentOutCode)
                    {
                        builder.Append("// ");
                    }
                    builder.Append(String.Format("\t\t{0}", t.typedef_signature));
                    builder.Append(Environment.NewLine);
                }
            }
            if (variables.Count > 0)
            {
                builder.Append(indent);
                builder.Append(String.Format("\t// Variables:"));
                builder.Append(Environment.NewLine);
                foreach (CVariable v in variables)
                {
                    builder.Append(indent);
                    if (commentOutCode)
                    {
                        builder.Append("// ");
                    }
                    builder.Append(String.Format("\t\t{0}", v.variable_signature));
                    builder.Append(Environment.NewLine);
                }
            }
            if (labels.Count > 0)
            {
                builder.Append(indent);
                builder.Append(String.Format("\t// Labels:"));
                builder.Append(Environment.NewLine);
                foreach (CLabel l in labels)
                {
                    builder.Append(indent);
                    builder.Append(String.Format("\t//\t\t0x{0:X8}\t{1}", l.offset_in_memory, l.name));
                    builder.Append(Environment.NewLine);
                }
            }

            int subBlockNum = 1;
            foreach (CBlock sub in sub_blocks)
            {
                builder.Append(Environment.NewLine);
                builder.Append(sub.GetStubCode(parentMethod, String.Format("{0}\t", indent), String.Format("{0}.{1}", blockName, subBlockNum), commentOutCode));
                subBlockNum++;
            }

            builder.Append(indent);
            builder.Append(String.Format("/* end block {0} */", blockName));
            builder.Append(Environment.NewLine);
            if (end_offset != 0)
            {
                builder.Append(indent);
                builder.Append(String.Format("// End offset: 0x{0:X8}", end_offset));
                builder.Append(Environment.NewLine);
            }
            builder.Append(indent);
            builder.Append(String.Format("// End Line: {0}", parentMethod.source_file_line_number + end_line));
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        public string ToGhidraLineNumberXML(string indent, CFunction parentFunction)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(indent);
            long lineNumber = start_line;
            if (lineNumber == 1)
            {
                if (parentFunction.source_file_line_number > 1)
                {
                    lineNumber = parentFunction.source_file_line_number;
                }
            }
            long endLineNumber = end_line;
            if (endLineNumber < lineNumber)
            {
                endLineNumber = lineNumber;
            }
            builder.Append(String.Format("<line_number source_file=\"{0}\" start=\"0x{1:X}\" end=\"0x{2:X}\" addr=\"0x{3:X}\" />", source_file, lineNumber, endLineNumber, start_offset));
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class CParameter : CElement
    {
        public string type_name { get; set; }
        public int register_number { get; set; }
        public string register_name { get; set; }
        public long stack_offset { get; set; }
        public string parameter_signature { get; set; }
        public bool is_function_pointer { get; set; }
        public bool is_varargs { get; set; }
        public string[] derived_types { get; set; }

        public CParameter()
            : base()
        {
            type_name = "";
            register_number = -1;
            //stack_offset = 0;
            parameter_signature = "";
            is_function_pointer = false;
            is_varargs = false;
            derived_types = new string[] { };
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public static string GetTypeNameForFunctionParameterReplacement(string typeName)
        {
            string result = typeName;
            result = result.Replace("/*stack ", "/*stack_");
            int lastChar = 0;
            int lastSpacePos = result.LastIndexOf(" ");
            if (lastSpacePos > 0)
            {
                lastChar = lastSpacePos;
            }
            int lastSlashPos = result.LastIndexOf("/");
            if ((lastSlashPos > 0) && (lastSlashPos < lastChar))
            {
                lastChar = lastSlashPos;
            }

            if (lastChar > 0)
            {
                result = result.Substring(0, lastChar).Trim();
            }

            return result;
        }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            //builder.Append(indent);
            //builder.Append(String.Format("<datatype name=\"{0}\" kind=\"Structure\" length=\"0x10\" >", name, primitive_type, lengthString));
            //builder.Append(Environment.NewLine);

            //builder.Append(indent);
            //builder.Append("</datatype>");
            //builder.Append(Environment.NewLine);

            return builder.ToString();
        }

        public string GetBarebonesParameterSignature()
        {
            string result = parameter_signature;
            if (result == null)
            {
                Console.WriteLine(string.Format("Warning: unexpected condition - null signature for parameter type '{0}', name '{1}'", type_name, name));
            }
            else
            {
                if (result.StartsWith("struct "))
                {
                    result = result.Substring(7);
                }
                if (result.StartsWith("union "))
                {
                    result = result.Substring(6);
                }
                if (result.StartsWith("enum "))
                {
                    result = result.Substring(5);
                }
            }
            return result;
        }
    }

    public class CFunction : CElement
    {
        public CParameter return_value { get; set; }
        public string stack_frame_base { get; set; }
        public uint stack_frame_size { get; set; }
        public uint mask { get; set; }
        public long mask_offset { get; set; }
        public List<string> saved_registers { get; set; }
        public List<CParameter> parameters { get; set; }
        public string method_signature { get; set; }
        public List<CBlock> blocks { get; set; }
        public string source_code { get; set; }

        public CFunction()
            : base()
        {
            saved_registers = new List<string>();
            parameters = new List<CParameter>();
            method_signature = "";
            blocks = new List<CBlock>();
            source_code = "";
        }

        //return $"{_returnType} /*${_register}*/ {_name}({string.Join(", ", _parameters)})";

        //public static CFunction FromSourceCode(string sourceCode)
        //{
        //    CFunction result = new CFunction();
        //    int codeStartIndex = sourceCode.IndexOf("{");
        //    int codeEndIndex = sourceCode.LastIndexOf("}");

        //    return result;
        //}

        //public static CFunction FromMethodSignature(string methodSignature)
        //{
        //    CFunction result = new CFunction();
        //    int paramStartIndex = methodSignature.IndexOf

        //    return result;
        //}

        public static string GetGhidrafiedRegisterName(CParameter param)
        {
            if (param.register_name == null)
            {
                return "";
            }
            string ghidraRegisterName = param.register_name;
            if (ghidraRegisterName.StartsWith("$"))
            {
                ghidraRegisterName = ghidraRegisterName.Substring(1);
            }
            // special cases
            if (ghidraRegisterName == "fp")
            {
                ghidraRegisterName = "s8";
                Console.WriteLine(string.Format("Info: Translating 'fp' (frame pointer) register reference into 's8' register reference for parameter with signature '{0}' for Ghidra compatibility.", param.parameter_signature));
            }
            return ghidraRegisterName;
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public bool HasFunctionPointerParameters()
        {
            if (return_value.is_function_pointer)
            {
                return true;
            }
            foreach (CParameter p in parameters)
            {
                if (p.is_function_pointer)
                {
                    return true;
                }
            }

            return false;
        }

        public override string GetHeaderCode()
        {
            StringBuilder builder = new StringBuilder();
            builder.Append(String.Format("extern {0}; // 0x{1:X8}", StripBlockCommentsFromString(method_signature), offset_in_memory));
            return builder.ToString();
        }

        public string GetFullCode()
        {
            //StringBuilder functionCodeBuilder = new StringBuilder();
            //functionCodeBuilder.Append(method_signature);
            //functionCodeBuilder.Append(Environment.NewLine);
            //functionCodeBuilder.Append("{");
            //functionCodeBuilder.Append(Environment.NewLine);
            //functionCodeBuilder.Append(source_code);
            //functionCodeBuilder.Append(Environment.NewLine);
            //functionCodeBuilder.Append("}");
            //functionCodeBuilder.Append(Environment.NewLine);
            //return functionCodeBuilder.ToString();
            return source_code;
        }

        public float MatchPercentWithOtherFunction(CFunction otherFunc, bool requireNameMatch)
        {
            float result = 0.0f;
            if (requireNameMatch)
            {
                if (name != otherFunc.name)
                {
                    return 0.0f;
                }
            }
            int totalNumParams = parameters.Count;
            int compareNumParams = parameters.Count;
            if (otherFunc.parameters.Count > totalNumParams)
            {
                totalNumParams = otherFunc.parameters.Count;
            }
            if (otherFunc.parameters.Count < compareNumParams)
            {
                compareNumParams = otherFunc.parameters.Count;
            }
            // add one for the return type
            totalNumParams++;
            float percentPerParam = 100.0f / (float)totalNumParams;
            // account for rounding errors
            int numDifferentParams = 0;
            if (return_value.type_name == otherFunc.return_value.type_name)
            {
                result += percentPerParam;
            }
            else
            {
                numDifferentParams++;
            }
            for (int i = 0; i < compareNumParams; i++)
            {
                if (parameters[i].type_name == otherFunc.parameters[i].type_name)
                {
                    result += percentPerParam;
                }
                else
                {
                    numDifferentParams++;
                }
            }
            if (numDifferentParams == 0)
            {
                result = 100.0f;
            }
            // last-chance fallback
            if ((result == 0.0f) && (name == otherFunc.name))
            {
                result = 1.0f;
            }

            return result;
        }

        public string GetBarebonesMethodSignature()
        {
            /* StringBuilder builder = new StringBuilder();
            builder.Append(return_value.GetBarebonesParameterSignature());
            builder.Append(" ");
            builder.Append(name);
            builder.Append("(");
            int paramNum = 0;
            foreach (CParameter p in parameters)
            {
                if (paramNum > 0)
                {
                    builder.Append(", ");
                }
                builder.Append(p.GetBarebonesParameterSignature());
                if (!p.type_name.EndsWith("*"))
                {
                    builder.Append(" ");
                }
                builder.Append(p.name);
                paramNum++;
            }
            builder.Append(")");
            return builder.ToString();*/
            // the point of this method was basically to strip out comments anyway, to make parsing the string easier.
            // so let's just do that for now and make it more complicated later if it needs to be.
            string result = method_signature;
            if (result.Contains("//"))
            {
                int eolCommentPos = result.IndexOf("//");
                result = result.Substring(0, eolCommentPos);
            }
            while (result.Contains("/*"))
            {
                int commentStart = result.IndexOf("/*");
                int commentEnd = result.IndexOf("*/", commentStart + 1);
                if (commentEnd < 0)
                {
                    result = result.Substring(0, commentStart);
                }
                else
                {
                    result = result.Substring(0, commentStart) + result.Substring(commentEnd + 2);
                }
            }
            while (result.Contains("  "))
            {
                result = result.Replace("  ", " ");
            }
            while (result.Contains(" ,"))
            {
                result = result.Replace(" ,", ",");
            }
            while (result.Contains(" )"))
            {
                result = result.Replace(" )", ")");
            }
            return result;
        }

        public override string GetSourceCode()
        {
            StringBuilder builder = new StringBuilder();
            bool commentOutStubCode = false;
            if ((source_code != null) && (source_code != ""))
            {
                commentOutStubCode = true;
                //builder.Append(method_signature);
                builder.Append("// decompiled code");
                builder.Append(Environment.NewLine);
                builder.Append("// original method signature: ");
                builder.Append(Environment.NewLine);
                builder.Append("// ");
                builder.Append(method_signature);
                builder.Append(Environment.NewLine);
            }
            else
            {
                builder.Append("// autogenerated function stub: ");
                builder.Append(Environment.NewLine);
                builder.Append("// ");
                builder.Append(method_signature);
                builder.Append(Environment.NewLine);
                builder.Append(GetBarebonesMethodSignature());
                builder.Append(Environment.NewLine);
                builder.Append("{");
            }
            builder.Append(String.Format(" // line {0}, offset 0x{1:x8}", source_file_line_number, offset_in_memory));
            builder.Append(Environment.NewLine);
            int blockNum = 1;
            if (blocks != null)
            {
                foreach (CBlock a in blocks)
                {
                    builder.Append(a.GetStubCode(this, "\t", String.Format("{0}", blockNum), commentOutStubCode));
                    builder.Append(Environment.NewLine);
                    blockNum++;
                }
            }
            string returnStub = "return null;";
            if (return_value != null)
            {
                switch (return_value.type_name)
                {
                    case "char":
                    case "unsigned char":
                    case "int":
                    case "unsigned int":
                    case "long":
                    case "unsigned long":
                    case "unsigned long long":
                    case "short":
                    case "unsigned short":
                        returnStub = "return 0;";
                        break;
                    case "void":
                        returnStub = "";
                        break;
                }
            }

            if ((source_code == null) || (source_code == ""))
                {
                if (returnStub != "")
                {
                    builder.Append(String.Format("\t{0}", returnStub));
                    builder.Append(Environment.NewLine);
                }
                builder.Append("}");
            }
            else
            {
                builder.Append(source_code);
                builder.Append(Environment.NewLine);
            }
            //builder.Append(String.Format(" // line {0}, offset {1}", SourceFileLineNumber, OriginalFunctionDef.offset_hex));
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }

        //public string GetFunctionStub(string skeletalPlaceholder)
        //{
        //    StringBuilder builder = new StringBuilder();
        //    bool commentOutStubCode = false;
        //    if (skeletalPlaceholder != "")
        //    {
        //        commentOutStubCode = true;
        //        //builder.Append(method_signature);
        //        builder.Append("// decompiled code");
        //        builder.Append(Environment.NewLine);
        //        builder.Append("// original method signature: ");
        //        builder.Append(Environment.NewLine);
        //        builder.Append("// ");
        //        builder.Append(method_signature);
        //        builder.Append(Environment.NewLine);
        //    }
        //    else
        //    {
        //        builder.Append("// autogenerated function stub: ");
        //        builder.Append(Environment.NewLine);
        //        builder.Append("// ");
        //        builder.Append(method_signature);
        //        builder.Append(Environment.NewLine);
        //        builder.Append(GetBarebonesMethodSignature());
        //        builder.Append(Environment.NewLine);
        //        builder.Append("{");
        //    }            
        //    builder.Append(String.Format(" // line {0}, offset 0x{1:x8}", source_file_line_number, offset_in_memory));
        //    builder.Append(Environment.NewLine);
        //    int blockNum = 1;
        //    if (blocks != null)
        //    {
        //        foreach (CBlock a in blocks)
        //        {
        //            builder.Append(a.GetStubCode(this, "\t", String.Format("{0}", blockNum), commentOutStubCode));
        //            builder.Append(Environment.NewLine);
        //            blockNum++;
        //        }
        //    }
        //    string returnStub = "return null;";
        //    if (return_value != null)
        //    {
        //        switch (return_value.type_name)
        //        {
        //            case "char":
        //            case "unsigned char":
        //            case "int":
        //            case "unsigned int":
        //            case "long":
        //            case "unsigned long":
        //            case "unsigned long long":
        //            case "short":
        //            case "unsigned short":
        //                returnStub = "return 0;";
        //                break;
        //            case "void":
        //                returnStub = "";
        //                break;
        //        }
        //    }

        //    if (skeletalPlaceholder == "")
        //    {
        //        if (returnStub != "")
        //        {
        //            builder.Append(String.Format("\t{0}", returnStub));
        //            builder.Append(Environment.NewLine);
        //        }
        //        builder.Append("}");
        //    }
        //    else
        //    {
        //        builder.Append(skeletalPlaceholder);
        //        builder.Append(Environment.NewLine);
        //    }
        //    //builder.Append(String.Format(" // line {0}, offset {1}", SourceFileLineNumber, OriginalFunctionDef.offset_hex));
        //    builder.Append(Environment.NewLine);
        //    return builder.ToString();
        //}

        public string ToGhidraFunctionXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(indent);
            // SYM data doesn't include the length of a function?
            //builder.Append(String.Format("<function name=\"{0}\" address=\"0x{1}\" length=\"0x{2}\">", name, offset, si));
            builder.Append(String.Format("<function name=\"{0}\" address=\"0x{1:X8}\">", name, offset_in_memory));
            builder.Append(Environment.NewLine);

            // <stack_variable name=\"%ws\" kind=\"%ws\" offset=\"0x%x\" datatype=\"%ws\" length=\"0x%x\" />
            //for (int i = 0; i < parameters.Count; i++)
            //{
            //    builder.Append(indent);
            //    builder.Append("\t");
            //    builder.Append(String.Format("<stack_variable name=\"{0}\" kind=\"{1}\" offset=\"0x{2:X}\" datatype=\"{3}\" length=\"0x0\" />", parameters[i].name, parameters[i].type_name, parameters[i].stack_offset, parameters[i].type_name));
            //    builder.Append(Environment.NewLine);
            //}

            for (int i = 0; i < blocks.Count; i++)
            {
                builder.Append(blocks[i].ToGhidraLineNumberXML(indent + "\t", this));
            }

            builder.Append(indent);
            builder.Append("</function>");
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }

        public string ToGhidraSymbolXML(string indent, int symbolIndex)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(indent);
            //builder.Append(String.Format("<symbol name=\"{0}\" address=\"0x{1:X}\" length=\"0x10\" tag=\"PublicSymbol\" kind=\"Unknown\" index=\"0x696d\" undecorated=\"bool __cdecl UIDataItems_MatchID(void*,void*)\" value=\"\" datatype=\"\" />", name, offset));
            builder.Append(String.Format("<symbol name=\"{0}\" address=\"0x{1:X}\" tag=\"Function\" kind=\"Unknown\" index=\"0x{2:X}\" undecorated=\"{3}\" value=\"\" datatype=\"{4}\" />", name, offset_in_memory, symbolIndex, method_signature, return_value.type_name));
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class MemorySegment
    {
        public string name { get; set; }
        public int segment_number { get; set; }
        public long offset { get; set; }
        public long length { get; set; }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(indent);
            builder.Append(String.Format("<segment number=\"{0}\" name=\"{1}\" address=\"0x{2:X8}\" length=\"0x{3:X8}\" />", segment_number, CElement.GhidraXMLEncodeString(name), offset, length));
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public struct AddressAndLineNumberTuple
    {
        public long address;
        public long line_number;
    }

    public struct PsyQSLDDataSet
    {
        public string source_file;
        public List<AddressAndLineNumberTuple> data;
        public int overlay_id;
    }

    public class PsyQOverlayDescription
    {
        public long overlay_id;
        public long offset;
        public long length;
        public string file_name;

        public List<CEnum> enums { get; set; }
        public List<CExternalDeclaration> external_declarations { get; set; }
        public List<CStruct> structs { get; set; }
        public List<CUnion> unions { get; set; }
        public List<CLabel> labels { get; set; }
        public List<CTypeDef> typedefs { get; set; }
        public List<CTypeDef> generated_typedefs { get; set; }
        public List<CFunction> functions { get; set; }
        public List<PsyQSLDDataSet> sld_data { get; set; }

        public PsyQOverlayDescription()
        {
            overlay_id = -1;
            offset = -1;
            length = -1;
            file_name = "";
            enums = new List<CEnum>();
            external_declarations = new List<CExternalDeclaration>();
            structs = new List<CStruct>();
            unions = new List<CUnion>();
            labels = new List<CLabel>();
            typedefs = new List<CTypeDef>();
            generated_typedefs = new List<CTypeDef>();
            functions = new List<CFunction>();
            sld_data = new List<PsyQSLDDataSet>();
        }

        public void AddFunctions(List<CFunction> newFunctions)
        {
            foreach (CFunction func in newFunctions)
            {
                functions.Add(func);
            }
        }

        public string GetDirectoryNameForDecompilation()
        {
            if (file_name != "")
            {
                return Utilities.GetStringAsSafeDirectoryName(file_name);
            }
            return string.Format("over0x{0:X2}", overlay_id);
        }

        public bool OverlapsWithOverlay(PsyQOverlayDescription otherOverlay)
        {
            if (offset == otherOverlay.offset)
            {
                return true;
            }
            if ((offset > otherOverlay.offset) && (offset < (otherOverlay.offset + otherOverlay.length)))
            {
                return true;
            }
            if ((offset < otherOverlay.offset) && ((offset + length) > otherOverlay.offset))
            {
                return true;
            }
            return false;
        }

        //Enums = new Dictionary<string, EnumDef>();
        //    Externs = new List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration>();
        //    Functions = new List<Function>();
        //    Labels = new Dictionary<uint, List<Label>>();
        //    Structs = new Dictionary<string, StructDef>();
        //    Typedefs = new Dictionary<string, TypeInfo>();
        //    Unions = new Dictionary<string, UnionDef>();
    }

    public class PsyQOverlayGroup
    {
        public long base_address;
        public List<PsyQOverlayDescription> overlays;

        public PsyQOverlayGroup()
        {
            base_address = -1;
            overlays = new List<PsyQOverlayDescription>();
        }

        public PsyQOverlayDescription GetOverlayByID(long overlayID)
        {
            foreach (PsyQOverlayDescription ol in overlays)
            {
                if (ol.overlay_id == overlayID)
                {
                    return ol;
                }
            }
            return null;
        }
    }

    public class ObjectFileData
    {
        public string file_name;
        public List<long> addresses;
        public long min_address;
        public long max_address;

        public ObjectFileData()
        {
            file_name = "";
            addresses = new List<long>();
            min_address = -1;
            max_address = -1;
        }
    }

    public class SymDumpSymbolCollection
    {
        public int version { get; set; }
        public int target_unit { get; set; }
        public List<MemorySegment> memory_segments { get; set; }
        public long program_counter { get; set; }
        public long global_pointer { get; set; }
        public long stack_pointer_base { get; set; }
        public long stack_pointer_size_or_offset { get; set; }

        public List<CEnum> enums { get; set; }
        public List<CExternalDeclaration> external_declarations { get; set; }
        public List<CStruct> structs { get; set; }
        public List<CUnion> unions { get; set; }
        public List<CLabel> labels { get; set; }
        public List<CTypeDef> typedefs { get; set; }
        public List<CTypeDef> generated_typedefs { get; set; }
        public List<CFunction> functions { get; set; }

        public List<PsyQSLDDataSet> sld_data { get; set; }

        public List<PsyQOverlayDescription> overlays { get; set; }

        public List<ObjectFileData> object_file_data { get; set; }

        public SymDumpSymbolCollection()
        {
            memory_segments = new List<MemorySegment>();
            overlays = new List<PsyQOverlayDescription>();
            sld_data = new List<PsyQSLDDataSet>();
            enums = new List<CEnum>();
            external_declarations = new List<CExternalDeclaration>();
            structs = new List<CStruct>();
            unions = new List<CUnion>();
            labels = new List<CLabel>();
            typedefs = new List<CTypeDef>();
            generated_typedefs = new List<CTypeDef>();
            functions = new List<CFunction>();
            object_file_data = new List<ObjectFileData>();
        }

        public PsyQOverlayDescription GetOverlayByID(long overlayID)
        {
            foreach (PsyQOverlayDescription ol in overlays)
            {
                if (ol.overlay_id == overlayID)
                {
                    return ol;
                }
            }
            return null;
        }

        public void SortAllElements()
        {
            enums.Sort();
            external_declarations.Sort();
            structs.Sort();
            unions.Sort();
            labels.Sort();
            typedefs.Sort();
            generated_typedefs.Sort();
            functions.Sort();
            foreach (PsyQOverlayDescription ol in overlays)
            {
                ol.enums.Sort();
                ol.external_declarations.Sort();
                ol.structs.Sort();
                ol.unions.Sort();
                ol.labels.Sort();
                ol.typedefs.Sort();
                ol.generated_typedefs.Sort();
                ol.functions.Sort();
            }
        }

        public string ToJSON()
        {
            return Newtonsoft.Json.JsonConvert.SerializeObject(this, Newtonsoft.Json.Formatting.Indented);
        }

        public void AddFunctions(List<CFunction> newFunctions)
        {
            foreach (CFunction func in newFunctions)
            {
                functions.Add(func);
            }
        }

        public static CLabel GetLabelByName(List<CLabel> labelList, string name)
        {
            CLabel result = null;

            foreach (CLabel l in labelList)
            {
                if (l.name == name)
                {
                    return l;
                }
            }

            return result;
        }

        public static List<CLabel> GetLabelsByAddress(List<CLabel> labelList, long address, int[] labelTypes)
        {
            List<CLabel> result = new List<CLabel>();

            foreach (CLabel l in labelList)
            {
                if (l.offset_in_memory == address)
                {
                    if ((labelTypes == null) || (labelTypes.Contains(l.label_type)))
                    {
                        result.Add(l);
                    }
                }
            }

            return result;
        }

        public string ToGhidraXML(string SYMFileName, string EXEFileName)
        {
            // This is done with per-class StringBuilders rather than some sort of 
            // XML serialization, because Ghidra's XML format is so specific
            StringBuilder builder = new StringBuilder();

            int symbolIndex = 2;

            List<string> allSourceFiles = new List<string>();

            // SYM files don't have GUIDs, so just randomly generate one
            Guid mg = Guid.NewGuid();
            string mockGUID = "{" + mg.ToString() + "}";

            // I don't know what an "age" of "6" means, it's just taken from a real Ghidra XML file
            builder.Append(String.Format("<pdb file=\"{0}\" exe=\"{1}\" guid=\"{2}\" age=\"6\">", CElement.GhidraXMLEncodeString(SYMFileName), CElement.GhidraXMLEncodeString(EXEFileName), mockGUID));
            builder.Append(Environment.NewLine);

            // enums
            builder.Append("\t<enums>");
            builder.Append(Environment.NewLine);

            foreach (CEnum def in enums)
            {
                builder.Append(def.ToGhidraXML("\t\t"));
                if (!allSourceFiles.Contains(def.source_file))
                {
                    allSourceFiles.Add(def.source_file);
                }
            }

            builder.Append("\t</enums>");
            builder.Append(Environment.NewLine);


            // datatypes (structs, unions, etc.)
            builder.Append("\t<datatypes>");
            builder.Append(Environment.NewLine);

            foreach (CStruct def in structs)
            {
                builder.Append(def.ToGhidraXML("\t\t"));
                if (!allSourceFiles.Contains(def.source_file))
                {
                    allSourceFiles.Add(def.source_file);
                }
            }

            foreach (CUnion def in unions)
            {
                builder.Append(def.ToGhidraXML("\t\t"));
                if (!allSourceFiles.Contains(def.source_file))
                {
                    allSourceFiles.Add(def.source_file);
                }
            }

            builder.Append("\t</datatypes>");
            builder.Append(Environment.NewLine);


            // typedefs
            builder.Append("\t<typedefs>");
            builder.Append(Environment.NewLine);

            foreach (CTypeDef def in typedefs)
            {
                builder.Append(def.ToGhidraXML("\t\t"));
                if (!allSourceFiles.Contains(def.source_file))
                {
                    allSourceFiles.Add(def.source_file);
                }
            }

            builder.Append("\t</typedefs>");
            builder.Append(Environment.NewLine);


            // classes
            builder.Append("\t<classes>");
            builder.Append(Environment.NewLine);

            builder.Append("\t</classes>");
            builder.Append(Environment.NewLine);


            // functions
            builder.Append("\t<functions>");
            builder.Append(Environment.NewLine);

            foreach (CFunction def in functions)
            {
                builder.Append(def.ToGhidraFunctionXML("\t\t"));
                if (!allSourceFiles.Contains(def.source_file))
                {
                    allSourceFiles.Add(def.source_file);
                }
            }

            builder.Append("\t</functions>");
            builder.Append(Environment.NewLine);


            // tables (begin)
            builder.Append("\t<tables>");
            builder.Append(Environment.NewLine);


            // source files
            builder.Append("\t\t<table name=\"SourceFiles\">");
            builder.Append(Environment.NewLine);

            allSourceFiles.Sort();

            foreach (string sf in allSourceFiles)
            {
                builder.Append(String.Format("\t\t\t<source_file name=\"{0}\" id=\"0x{1:X}\" />", sf, symbolIndex));
                builder.Append(Environment.NewLine);
                symbolIndex++;
            }

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // line numbers
            builder.Append("\t\t<table name=\"LineNumbers\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // sections
            builder.Append("\t\t<table name=\"Sections\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // memory segment map
            builder.Append("\t\t<table name=\"SegmentMap\">");
            builder.Append(Environment.NewLine);

            foreach (MemorySegment def in memory_segments)
            {
                builder.Append(def.ToGhidraXML("\t\t\t"));
            }

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // injected source - don't know if there's an equivalent in a SYM file
            builder.Append("\t\t<table name=\"InjectedSource\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // frame data - don't know if there's an equivalent in a SYM file
            builder.Append("\t\t<table name=\"FrameData\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // input assembly files - don't know if there's an equivalent in a SYM file
            builder.Append("\t\t<table name=\"InputAssemblyFiles\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // symbols - not 100% sure how to handle this yet
            // probably have to dynamically create it from all the other data
            builder.Append("\t\t<table name=\"Symbols\">");
            builder.Append(Environment.NewLine);



            builder.Append(String.Format("\t\t\t<symbol name=\"{0}\" address=\"0x0\" length=\"0x0\" tag=\"Executable\" kind=\"Unknown\" index=\"0x1\" undecorated=\"\" value=\"\" datatype=\"\"  />", EXEFileName));
            builder.Append(Environment.NewLine);

            // "Data" tag - labels?
            // <symbol name="__lc_codepage" address="0xf9d6e8" length="0x0" tag="Data" kind="Global" index="0x8f2" undecorated="" value="" datatype="uint"  />

            foreach (CFunction def in functions)
            {
                builder.Append(def.ToGhidraSymbolXML("\t\t\t", symbolIndex));
                symbolIndex++;
            }

            foreach (CLabel def in labels)
            {
                builder.Append(def.ToGhidraSymbolXML("\t\t\t", symbolIndex));
                symbolIndex++;
            }

            // <symbol name="key" address="0x0" length="0x0" tag="Data" kind="Parameter" index="0x6a04" undecorated="" value="" datatype="cdc::SceneId *"  />

            // <symbol name="last" address="0x0" length="0x0" tag="Data" kind="Local" index="0x6a05" undecorated="" value="" datatype="uint"  />



            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);

            builder.Append("\t</tables>");
            builder.Append(Environment.NewLine);
            // tables (end)

            builder.Append("</pdb>");
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }
    }
}
