﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace BenLincoln.ReverseEngineeringUtils
{
    public class EnumDef
    {
        public string name { get; set; }
        public string source_file { get; set; }
        public string primitive_type { get; set; }
        public Dictionary<string, string> members { get; set; }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            string lengthString = "0x4";
            if (primitive_type != "int")
            {
                throw new Exception(String.Format("Unsupported enum primitive type: {0}", primitive_type));
            }

            builder.Append(indent);
            builder.Append(String.Format("<enum name=\"{0}\" type=\"{1}\" length=\"{2}\">", name, primitive_type, lengthString));
            builder.Append(Environment.NewLine);

            foreach (string memberName in members.Keys)
            {
                builder.Append(indent);
                builder.Append("\t");
                int memberVal = -1;
                try
                {
                    memberVal = int.Parse(members[memberName]);
                }
                catch (Exception ex)
                {
                    throw new Exception(String.Format("Could not parse '{0}' as an integer. Is the type supported?", members[memberName]));
                }
                string memberValFormatted = String.Format("0x{0:x}", memberVal);
                builder.Append(String.Format("<member name=\"{0}\" value=\"{1}\" />", memberName, memberValFormatted));
                builder.Append(Environment.NewLine);
            }

            builder.Append(indent);
            builder.Append("</enum>");
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class UnionDef
    {
        public string name { get; set; }
        public string source_file { get; set; }
        public List<string> members { get; set; }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            return builder.ToString();
        }
    }

    public class StructDef
    {
        public string name { get; set; }
        public string source_file { get; set; }
        public List<string> members { get; set; }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            builder.Append(indent);
            builder.Append(String.Format("<datatype name=\"{0}\" kind=\"Structure\" length=\"0x10\" >", name, primitive_type, lengthString));
            builder.Append(Environment.NewLine);

            builder.Append(indent);
            builder.Append("</datatype>");
            builder.Append(Environment.NewLine);

            return builder.ToString();
        }
    }

    public class TypeDef
    {
        public string name { get; set; }
        public string source_file { get; set; }
        public string definition { get; set; }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            return builder.ToString();
        }
    }

    public class LabelDef
    {
        public string name { get; set; }
        public string offset_dec { get; set; }
        public string offset_hex { get; set; }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            return builder.ToString();
        }
    }

    public class ExternalDeclarationDef
    {
        public string source_file { get; set; }
        public string declaration { get; set; }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            return builder.ToString();
        }
    }

    public class BlockDef
    {
        public string source_file { get; set; }
        public string start_line { get; set; }
        public string end_line { get; set; }
        public string start_offset { get; set; }
        public string end_offset { get; set; }
        //public List<TypeDef> typedefs { get; set; }
        public List<string> typedefs { get; set; }
        public List<string> variables { get; set; }
        public List<LabelDef> labels { get; set; }
        public List<BlockDef> sub_blocks { get; set; }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            return builder.ToString();
        }
    }

    public class FunctionDef
    {
        public string name { get; set; }
        public string source_file { get; set; }
        public string source_file_line_number { get; set; }
        public string offset { get; set; }
        public string stack_frame_base { get; set; }
        public string stack_frame_size { get; set; }
        public string mask_offset { get; set; }
        public List<string> saved_registers { get; set; }
        public string method_signature { get; set; }
        public List<BlockDef> blocks { get; set; }

        public string ToGhidraXML(string indent)
        {
            StringBuilder builder = new StringBuilder();

            return builder.ToString();
        }
    }

    public class SymDumpJSON
    {
        public int Version { get; set; }
        public int targetUnit { get; set; }
        public List<EnumDef> enums { get; set; }
        public List<UnionDef> unions { get; set; }
        public List<StructDef> structs { get; set; }
        public List<LabelDef> labels { get; set; }
        public List<TypeDef> typedefs { get; set; }
        public List<ExternalDeclarationDef> external_declarations { get; set; }
        public List<FunctionDef> functions { get; set; }

        public static string GhidraXMLEncodeString(string InputString)
        {
            string result = InputString;

            result = result.Replace("&", "&amp;");
            result = result.Replace("'", "&apos;");
            result = result.Replace("<", "&lt;");
            result = result.Replace(">", "&gt;");
            result = result.Replace("\"", "&quot;");

            return result;
        }

        public string ToGhidraXML(string SYMFileName, string EXEFileName)
        {
            // This is done with per-class StringBuilders rather than some sort of 
            // XML serialization, because Ghidra's XML format is so specific
            StringBuilder builder = new StringBuilder();

            // SYM files don't have GUIDs, so just randomly generate one
            Guid mg = Guid.NewGuid();
            string mockGUID = "{" + mg.ToString() + "}";

            // I don't know what an "age" of "6" means, it's just taken from a real Ghidra XML file
            builder.Append(String.Format("<pdb file=\"{0}\" exe=\"{1}\" guid=\"{2}\" age=\"6\">", SYMFileName, EXEFileName, mockGUID));
            builder.Append(Environment.NewLine);

            // enums
            builder.Append("\t<enums>");
            builder.Append(Environment.NewLine);

            foreach (EnumDef def in enums)
            {
                builder.Append(def.ToGhidraXML("\t\t"));
            }

            builder.Append("\t</enums>");
            builder.Append(Environment.NewLine);


            // datatypes (structs, unions, etc.)
            builder.Append("\t<datatypes>");
            builder.Append(Environment.NewLine);

            foreach (StructDef def in structs)
            {
                builder.Append(def.ToGhidraXML("\t\t"));
            }

            foreach (UnionDef def in unions)
            {
                builder.Append(def.ToGhidraXML("\t\t"));
            }

            builder.Append("\t</datatypes>");
            builder.Append(Environment.NewLine);


            // typedefs
            builder.Append("\t<typedefs>");
            builder.Append(Environment.NewLine);

            builder.Append("\t</typedefs>");
            builder.Append(Environment.NewLine);


            // classes
            builder.Append("\t<classes>");
            builder.Append(Environment.NewLine);

            builder.Append("\t</classes>");
            builder.Append(Environment.NewLine);


            // functions
            builder.Append("\t<functions>");
            builder.Append(Environment.NewLine);

            builder.Append("\t</functions>");
            builder.Append(Environment.NewLine);


            // tables (begin)
            builder.Append("\t<tables>");
            builder.Append(Environment.NewLine);


            // source files
            builder.Append("\t\t<table name=\"SourceFiles\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // line numbers
            builder.Append("\t\t<table name=\"LineNumbers\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // sections
            builder.Append("\t\t<table name=\"Sections\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // segment map - building this from the PSX memory layout
            builder.Append("\t\t<table name=\"SegmentMap\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // injected source - don't know if there's an equivalent in a SYM file
            builder.Append("\t\t<table name=\"InjectedSource\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // frame data - don't know if there's an equivalent in a SYM file
            builder.Append("\t\t<table name=\"FrameData\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // input assembly files - don't know if there's an equivalent in a SYM file
            builder.Append("\t\t<table name=\"InputAssemblyFiles\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);


            // symbols - not 100% sure how to handle this yet
            // probably have to dynamically create it from all the other data
            builder.Append("\t\t<table name=\"Symbols\">");
            builder.Append(Environment.NewLine);

            builder.Append("\t\t</table>");
            builder.Append(Environment.NewLine);

            builder.Append("\t</tables>");
            builder.Append(Environment.NewLine);
            // tables (end)

            builder.Append("</pdb>");
            builder.Append(Environment.NewLine);
            return builder.ToString();
        }
    }
}
