﻿// PlayStation ELF Converter
// Converts between ELF and PSX-EXE executable formats
// Copyright 2019 Ben Lincoln
// https://www.beneaththewaves.net/
// Based in part on information documented in LameGuy64's elf2x utility
// (http://www.psxdev.net/forum/viewtopic.php?f=75&t=1613)

// This file is part of PlayStation ELF Converter.

// PlayStation ELF Converter is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// PlayStation ELF Converter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with PlayStation ELF Converter (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using BenLincoln.DebugSymbols;

namespace BenLincoln.ExecutableFiles
{
    public class MemorySegment
    {
        public string Name { get; set; }
        public int SegmentNumber { get; set; }
        public uint Offset { get; set; }
        public uint Length { get; set; }
        public byte[] Data { get; set; }

        public MemorySegment()
        {
            Name = "";
            SegmentNumber = -1;
            Offset = 0;
            Length = 0;
        }

        public static MemorySegment FromCommandLineParams(string commandLineParams)
        {
            // --overlay "[ID],[OFFSET],[LENGTH],[SOURCE_FILE](,[START_OFFSET],[LENGTH])"
            MemorySegment result = new MemorySegment();
            string[] paramArray = commandLineParams.Split(new string[] { "," }, StringSplitOptions.None);
            if (paramArray.Length < 4)
            {
                throw new Exception(string.Format("Not enough parameters in '{0}' to define a memory segment", commandLineParams));
            }
            result.SegmentNumber = (int)BenLincoln.DebugSymbols.Utilities.GetHexOrDecimalLong(paramArray[0]);
            result.Offset = (uint)BenLincoln.DebugSymbols.Utilities.GetHexOrDecimalLong(paramArray[1]);
            result.Length = (uint)BenLincoln.DebugSymbols.Utilities.GetHexOrDecimalLong(paramArray[2]);
            result.Name = string.Format("overlay0x{0:X2}", result.SegmentNumber);
            string sourceFile = paramArray[3];
            long sourceFileOffset = 0;
            long sourceFileLength = -1;
            if (paramArray.Length > 4)
            {
                sourceFileOffset = BenLincoln.DebugSymbols.Utilities.GetHexOrDecimalLong(paramArray[4]);
            }
            if (paramArray.Length > 5)
            {
                sourceFileOffset = BenLincoln.DebugSymbols.Utilities.GetHexOrDecimalLong(paramArray[5]);
            }

            FileStream fs = new FileStream(sourceFile, FileMode.Open, FileAccess.Read);
            BinaryReader r = new BinaryReader(fs);
            if (sourceFileOffset > 0)
            {
                fs.Seek(sourceFileOffset, SeekOrigin.Begin);
            }

            byte[] data = BenLincoln.DebugSymbols.Utilities.ReadAllBytes(r);

            result.Data = new byte[result.Length];

            long numBytes = result.Length;

            if (sourceFileLength > 0)
            {
                if (sourceFileLength < numBytes)
                {
                    numBytes = sourceFileLength;
                }
                if (data.Length < numBytes)
                {
                    numBytes = data.Length;
                }
            }
            Array.Copy(data, result.Data, numBytes);
            if (numBytes < result.Length)
            {
                for (long bNum = numBytes; numBytes < result.Length; numBytes++)
                {
                    result.Data[bNum] = 0;
                }
            }

            r.Close();
            fs.Close();

            return result;
        }
    }
}
