using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;
using symdump.symfile.util;
using symdump.util;

namespace symdump.symfile
{
    public class Block
    {
        private readonly uint _endLine;
        private readonly uint _endOffset;

        private readonly Function _function;
        private readonly List<Label> _labels = new List<Label>();
        private readonly uint _startLine;

        private readonly uint _startOffset;
        private readonly List<Block> _subBlocks = new List<Block>();
        private readonly Dictionary<string, TypeInfo> _typedefs = new Dictionary<string, TypeInfo>();
        private readonly List<BenLincoln.DebugSymbols.SymDump.CVariable> _vars = new List<BenLincoln.DebugSymbols.SymDump.CVariable>();
        public readonly int SYMFileSequenceNumber;

        public Block(BinaryReader reader, uint ofs, uint ln, Function f, SymDumpProcessingConfiguration config)
        {
            _startOffset = ofs;
            _startLine = ln;
            _function = f;
            SYMFileSequenceNumber = config.Issuer.GetNextNumber();

            while (true)
            {
                var typedValue = new TypedValue(reader, config);

                if (reader.SkipSld(typedValue))
                    continue;

                switch (typedValue.Type & 0x7f)
                {
                    case 16:
                        _subBlocks.Add(new Block(reader, (uint) typedValue.Value, reader.ReadUInt32(), _function, config));
                        break;
                    case 18:
                        _endOffset = (uint) typedValue.Value;
                        _endLine = reader.ReadUInt32();
                        return;
                    case 20:
                    {
                        var ti = reader.ReadTypeInfo(false, config);
                        var memberName = reader.ReadPascalString();
                        Debug.Assert(!string.IsNullOrEmpty(memberName));
                        switch (ti.ClassType)
                        {
                            case ClassType.AutoVar:
                                //_vars.Add($"{ti.AsCode(memberName)}; // stack offset {typedValue.Value}");
                                BenLincoln.DebugSymbols.SymDump.CVariable v = new BenLincoln.DebugSymbols.SymDump.CVariable();
                                v.name = memberName;
                                v.variable_type = "stack";
                                v.variable_signature = $"{ti.AsCode(memberName)}; // stack offset {typedValue.Value}";
                                v.stack_offset = typedValue.Value;
                                //v.c_type = ti.GetCType();
                                //v.type_name = ti.GetName(memberName);
                                //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed1 = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(v.variable_signature, true);
                                //v.c_type = pParsed1[0].type_name;
                                //v.type_name = pParsed1[0].type_name;
                                v.c_type = ti.AsCode("");
                                v.type_name = v.c_type;
                                v.size = ti.Size;
                                _vars.Add(v);
                                break;
                            case ClassType.Register:
                                //_vars.Add($"{ti.AsCode(memberName)}; // ${(Register) typedValue.Value}");
                                BenLincoln.DebugSymbols.SymDump.CVariable vr = new BenLincoln.DebugSymbols.SymDump.CVariable();
                                vr.name = memberName;
                                vr.variable_type = "register";
                                vr.variable_signature = $"{ti.AsCode(memberName)}; // ${(Register)typedValue.Value}";
                                vr.register_name = ((Register)typedValue.Value).ToString();
                                //vr.c_type = ti.GetCType();
                                //vr.type_name = ti.GetName(memberName);
                                //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed2 = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(vr.variable_signature, true);
                                //vr.c_type = pParsed2[0].type_name;
                                //vr.type_name = pParsed2[0].type_name;
                                vr.c_type = ti.AsCode("");
                                vr.type_name = vr.c_type;
                                vr.size = ti.Size;
                                _vars.Add(vr);
                                break;
                            case ClassType.Static:
                                //_vars.Add($"static {ti.AsCode(memberName)}; // offset 0x{typedValue.Value:x}");
                                BenLincoln.DebugSymbols.SymDump.CVariable vs = new BenLincoln.DebugSymbols.SymDump.CVariable();
                                vs.name = memberName;
                                vs.variable_type = "static";
                                vs.variable_signature = $"static {ti.AsCode(memberName)}; // offset 0x{typedValue.Value:x}";
                                vs.static_offset = (uint)typedValue.Value;
                                //vs.c_type = ti.GetCType();
                                //vs.type_name = ti.GetName(memberName);
                                //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed3 = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(vs.variable_signature, true);
                                //vs.c_type = pParsed3[0].type_name;
                                //vs.type_name = pParsed3[0].type_name;
                                vs.c_type = ti.AsCode("");
                                vs.type_name = vs.c_type;
                                vs.size = ti.Size;
                                _vars.Add(vs);
                                break;
                            case ClassType.Typedef:
                                _typedefs.Add(memberName, ti);
                                break;
                            case ClassType.Label:
                                _labels.Add(new Label(typedValue, memberName));
                                break;
                            default:
                                throw new Exception($"Unexpected class type {ti.ClassType}");
                        }

                        break;
                    }
                    case 22:
                    {
                        var ti = reader.ReadTypeInfo(true, config);
                        var memberName = reader.ReadPascalString();
                        Debug.Assert(!string.IsNullOrEmpty(memberName));
                        switch (ti.ClassType)
                        {
                            case ClassType.AutoVar:
                                //_vars.Add($"{ti.AsCode(memberName)}; // stack offset {typedValue.Value}");
                                BenLincoln.DebugSymbols.SymDump.CVariable v = new BenLincoln.DebugSymbols.SymDump.CVariable();
                                v.name = memberName;
                                v.variable_type = "stack";
                                v.variable_signature = $"{ti.AsCode(memberName)}; // stack offset {typedValue.Value}";
                                v.stack_offset = typedValue.Value;
                                //v.c_type = ti.GetCType();
                                //v.type_name = ti.GetName(memberName);
                                //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed4 = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(v.variable_signature, true);
                                //v.c_type = pParsed4[0].type_name;
                                //v.type_name = pParsed4[0].type_name;
                                v.c_type = ti.AsCode("");
                                v.type_name = v.c_type;
                                v.size = ti.Size;
                                _vars.Add(v);
                                break;
                            case ClassType.Register:
                                //_vars.Add($"{ti.AsCode(memberName)}; // ${(Register) typedValue.Value}");
                                BenLincoln.DebugSymbols.SymDump.CVariable vr = new BenLincoln.DebugSymbols.SymDump.CVariable();
                                vr.name = memberName;
                                vr.variable_type = "register";
                                vr.variable_signature = $"{ti.AsCode(memberName)}; // ${(Register)typedValue.Value}";
                                vr.register_name = ((Register)typedValue.Value).ToString();
                                //vr.c_type = ti.GetCType();
                                //vr.type_name = ti.GetName(memberName);
                                //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed5 = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(vr.variable_signature, true);
                                //vr.c_type = pParsed5[0].type_name;
                                //vr.type_name = pParsed5[0].type_name;
                                vr.c_type = ti.AsCode("");
                                vr.type_name = vr.c_type;
                                vr.size = ti.Size;
                                _vars.Add(vr);
                                break;
                            case ClassType.Static:
                                //_vars.Add($"static {ti.AsCode(memberName)}; // offset 0x{typedValue.Value:x}");
                                BenLincoln.DebugSymbols.SymDump.CVariable vs = new BenLincoln.DebugSymbols.SymDump.CVariable();
                                vs.name = memberName;
                                vs.variable_type = "static";
                                vs.variable_signature = $"static {ti.AsCode(memberName)}; // offset 0x{typedValue.Value:x}";
                                vs.static_offset = (uint)typedValue.Value;
                                //vs.c_type = ti.GetCType();
                                //vs.type_name = ti.GetName(memberName);
                                //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed6 = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(vs.variable_signature, true);
                                //vs.c_type = pParsed6[0].type_name;
                                //vs.type_name = pParsed6[0].type_name;
                                vs.c_type = ti.AsCode("");
                                vs.type_name = vs.c_type;
                                vs.size = ti.Size;
                                _vars.Add(vs);
                                break;
                            case ClassType.Typedef:
                                _typedefs.Add(memberName, ti);
                                break;
                            case ClassType.Label:
                                _labels.Add(new Label(typedValue, memberName));
                                break;
                            default:
                                throw new Exception($"Unexpected class type {ti.ClassType}");
                        }

                        break;
                    }
                }
            }
        }

        public void Dump(IndentedTextWriter writer)
        {
            writer.WriteLine($"{{ // line {_startLine}, offset 0x{_startOffset:x}");
            ++writer.Indent;
            foreach (var t in _typedefs)
                writer.WriteLine($"typedef {t.Value.AsCode(t.Key)};");
            _vars.ForEach(writer.WriteLine);
            foreach (var l in _labels)
                writer.WriteLine(l);
            _subBlocks.ForEach(b => b.Dump(writer));
            --writer.Indent;
            writer.WriteLine($"}} // line {_endLine}, offset 0x{_endOffset:x}");
        }

        public BenLincoln.DebugSymbols.SymDump.CBlock ToBLDSCBlock(string sourceFile)
        {
            BenLincoln.DebugSymbols.SymDump.CBlock result = new BenLincoln.DebugSymbols.SymDump.CBlock();
            result.source_file = sourceFile;
            result.start_line = _startLine;
            result.end_line = _endLine;
            result.start_offset = _startOffset;
            result.end_offset = _endOffset;
            result.hashcode = GetHashCode();
            result.typedefs = new List<BenLincoln.DebugSymbols.SymDump.CTypeDef>();
            result.sym_file_sequence_number = SYMFileSequenceNumber;
            foreach (string tiName in _typedefs.Keys)
            {
                TypeInfo ti = _typedefs[tiName];
                result.typedefs.Add(ti.ToBLDSCTypeDef(tiName, sourceFile));
            }
            result.variables = _vars;
            result.labels = new List<BenLincoln.DebugSymbols.SymDump.CLabel>();
            foreach (Label l in _labels)
            {
                result.labels.Add(l.ToBLDSCLabel());
            }
            result.sub_blocks = new List<BenLincoln.DebugSymbols.SymDump.CBlock>();
            foreach (Block b in _subBlocks)
            {
                result.sub_blocks.Add(b.ToBLDSCBlock(sourceFile));
            }

            return result;
        }

        public string DumpJSON(string sourceFile, string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CBlock resultDef = ToBLDSCBlock(sourceFile);

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);

            //StringBuilder builder = new StringBuilder();

            //builder.Append(lineIndent);
            //builder.Append("{");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"source_file\": \"{0}\",", sourceFile));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"start_line\": \"{0}\",", _startLine));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"end_line\": \"{0}\",", _endLine));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"start_offset\": \"{0}\",", _startOffset));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"end_offset\": \"{0}\",", _endOffset));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("\t\"typedefs\": ");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("\t[");
            //builder.Append(Environment.NewLine);
            //int elemNum = 1;
            //foreach (var t in _typedefs)
            //{
            //    builder.Append(lineIndent);
            //    builder.Append(String.Format("\t\t\"{0}\"", t.Value.AsCode(t.Key)));
            //    if (elemNum < _typedefs.Count)
            //    {
            //        builder.Append(",");
            //    }
            //    elemNum++;
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append("\t],");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("\t\"variables\": ");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("\t[");
            //builder.Append(Environment.NewLine);
            //elemNum = 1;
            //foreach (var v in _vars)
            //{
            //    builder.Append(lineIndent);
            //    builder.Append(String.Format("\t\t\"{0}\"", v));
            //    if (elemNum < _vars.Count)
            //    {
            //        builder.Append(",");
            //    }
            //    elemNum++;
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append("\t],");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("\t\"labels\": ");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("\t[");
            //builder.Append(Environment.NewLine);
            //elemNum = 1;
            //foreach (var l in _labels)
            //{
            //    string newIndent = String.Format("{0}\t\t", lineIndent);
            //    builder.Append(l.DumpJSON(newIndent));
            //    if (elemNum < _labels.Count)
            //    {
            //        builder.Append(",");
            //    }
            //    elemNum++;
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append("\t],");
            //builder.Append(Environment.NewLine);

            //if (_subBlocks.Count != 0)
            //{
            //    builder.Append(lineIndent);
            //    builder.Append("\t\"sub_blocks\": ");
            //    builder.Append(Environment.NewLine);
            //    builder.Append(lineIndent);
            //    builder.Append("\t[");
            //    builder.Append(Environment.NewLine);
            //    string newIndent = String.Format("{0}\t\t", lineIndent);
            //    int blockNum = 1;
            //    foreach (Block b in _subBlocks)
            //    {
            //        builder.Append(b.DumpJSON(sourceFile, newIndent));
            //        if (blockNum < _subBlocks.Count)
            //        {
            //            builder.Append(",");
            //        }
            //        blockNum++;
            //        builder.Append(Environment.NewLine);
            //    }
            //    builder.Append(lineIndent);
            //    builder.Append("\t]");
            //    builder.Append(Environment.NewLine);
            //}
            //builder.Append(lineIndent);
            //builder.Append("}");

            //return builder.ToString();
        }
    }
}
