﻿using System;
using System.IO;
using System.Text;
using symdump.symfile.util;

namespace symdump.symfile
{
    public enum LabelType
    {
        Offset = 1,
        Function2 = 2,
        Function6 = 6
    }

    public class Label
    {
        private readonly TypedValue _typedOffset;
        public string SLDSourceFileName;
        public string ObjectFileName;

        public Label(TypedValue typedValue, BinaryReader fs)
        {
            _typedOffset = typedValue;
            Name = fs.ReadPascalString();
        }

        public Label(TypedValue typedValue, string name)
        {
            _typedOffset = typedValue;
            Name = name;
        }

        public uint Offset => (uint) _typedOffset.Value;

        public byte LabelType => (byte)_typedOffset.Type;

        public string Name { get; }

        public override string ToString()
        {
            return $"0x{Offset:X8} {Name} // type: 0x{LabelType:X2}, object file: '{ObjectFileName}'";
        }

        public BenLincoln.DebugSymbols.SymDump.CLabel ToBLDSCLabel()
        {
            BenLincoln.DebugSymbols.SymDump.CLabel result = new BenLincoln.DebugSymbols.SymDump.CLabel();

            result.name = Name;
            result.hashcode = GetHashCode();
            result.offset_in_memory = Offset;
            result.sym_file_sequence_number = _typedOffset.SYMFileSequenceNumber;
            result.source_file = SLDSourceFileName;
            result.object_files.Add(ObjectFileName);

            return result;
        }

        public string DumpJSON(string lineIndent)
        {
            BenLincoln.DebugSymbols.SymDump.CLabel resultDef = ToBLDSCLabel();

            return BenLincoln.DebugSymbols.Utilities.IndentString(resultDef.ToJSON(), lineIndent);

            //StringBuilder builder = new StringBuilder();

            //builder.Append(lineIndent);
            //builder.Append("{");
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"name\": \"{0}\",", Name));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"offset_dec\": \"{0}\",", Offset));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append(String.Format("\t\"offset_hex\": \"0x{0:X8}\",", Offset));
            //builder.Append(Environment.NewLine);
            //builder.Append(lineIndent);
            //builder.Append("}");

            //return builder.ToString();
        }
    }
}
