﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;
using symdump.symfile.util;

namespace symdump.symfile
{
    public class StructMember : IEquatable<StructMember>
    {
        public string name;
        public readonly TypedValue TypedValue;
        public TypeInfo TypeInfo;
        public string ParentName;
        public readonly int ParentHashCode;
        public bool OriginalWasFake;
        protected bool _InlineFakes;
        public List<string> ObjectFileNames;
        // there are less messy ways to do this, but they're a lot more work, and this seems like an edge case
        protected StructDef _InlineStructDef;
        protected UnionDef _InlineUnionDef;
        protected SymDumpProcessingConfiguration ProcessingConfig;

        public StructMember(TypedValue tv, BinaryReader reader, bool extended, string parentName, int parentHashCode,
            SymDumpProcessingConfiguration config)
        {
            ProcessingConfig = config;
            TypeInfo = reader.ReadTypeInfo(extended, config);
            name = reader.ReadPascalString();

            ObjectFileNames = new List<string>();
            ObjectFileNames.Add(config.Tracker.CurrentObjectFileData.FileName);

            OriginalWasFake = IsFake();

            TypedValue = tv;
            ParentName = parentName;
            ParentHashCode = parentHashCode;
            _InlineFakes = config.InlineFakes;

            if (config.RenameForCompatibility)
            {
                name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(name, true);
            }

            if (name == ".eos")
            {
                OriginalWasFake = false;
            }

            //if (name.Contains("fake"))
            //{
            //    Console.WriteLine("Debug");
            //}

            //if (OriginalWasFake)
            //{
            //    Console.WriteLine("Debug");
            //}

            //if (config.AutoRenameFakes && (OriginalWasFake))
            //{
            //    string lastObjectFileName = "";
            //    if (config.Tracker.ObjectFileDataEntries.Count > 0)
            //    {
            //        lastObjectFileName = config.Tracker.ObjectFileDataEntries[config.Tracker.ObjectFileDataEntries.Count - 1].FileName;
            //    }
            //    string lastSLDSourceFileName = "";
            //    if (config.Tracker.SLDEntries.Count > 0)
            //    {
            //        lastSLDSourceFileName = config.Tracker.SLDEntries[config.Tracker.SLDEntries.Count - 1].SLDFile;
            //    }

            //    string cTypeBase = TypeInfo.AsCode("");
            //    string cType = "unknown";
            //    string cSubType = "unknown";
            //    if (cTypeBase.Contains(" "))
            //    {
            //        string[] cTypeArray = cTypeBase.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
            //        cType = cTypeArray[0];
            //        if (cTypeArray.Length > 1)
            //        {
            //            cSubType = cTypeArray[1];
            //        }
            //    }
            //    string parentType = "unknown";
            //    if (TypeInfo.ClassType == ClassType.StructMember)
            //    {
            //        parentType = "struct";
            //    }
            //    if (TypeInfo.ClassType == ClassType.UnionMember)
            //    {
            //        parentType = "union";
            //    }
            //    List<string> cSubTypes = new List<string>();
            //    cSubTypes.Add(cSubType);
            //    // get the original fake name of the parent
            //    string originalCSubType = cSubType;
            //    int ulPos = cSubType.LastIndexOf("_");
            //    if (ulPos > 0)
            //    {
            //        originalCSubType = originalCSubType.Substring(ulPos);
            //        cSubTypes.Add(originalCSubType);
            //    }
            //    bool foundMatch = false;
            //    if (cType == "struct")
            //    {
            //        for (int i = 0; i < cSubTypes.Count; i++)
            //        {
            //            if (config.FakeStructs.ContainsKey(cSubTypes[i]))
            //            {
            //                _InlineStructDef = config.FakeStructs[cSubTypes[i]];
            //                SymFile.OutputDebugMessage(string.Format("Matched fake {0} '{1} {2}' in {3} '{4}' with existing fake '{5}'", 
            //                    cType, cSubType, name, parentType, parentName, cSubTypes[i]), config);
            //                config.FakeStructs[cSubTypes[i]].HasBeenInlined = true;
            //                foundMatch = true;
            //                break;
            //            }
            //        }
            //    }
            //    if (cType == "union")
            //    {
            //        for (int i = 0; i < cSubTypes.Count; i++)
            //        {
            //            if (config.FakeUnions.ContainsKey(cSubTypes[i]))
            //            {
            //                _InlineUnionDef = config.FakeUnions[cSubTypes[i]];
            //                SymFile.OutputDebugMessage(string.Format("Matched fake {0} '{1} {2}' in {3} '{4}' with existing fake '{5}'",
            //                    cType, cSubType, name, parentType, parentName, cSubTypes[i]), config);
            //                config.FakeUnions[cSubTypes[i]].HasBeenInlined = true;
            //                foundMatch = true;
            //                break;
            //            }
            //        }
            //    }
            //    if (!foundMatch)
            //    {
            //        SymFile.OutputDebugMessage(string.Format("Error: did not find fake {0} '{1}' when populating {1} '{2}'", cType, cSubTypes, parentType, parentName), config);
            //    }
            //}

            if (TypeInfo.ClassType != ClassType.Bitfield && TypeInfo.ClassType != ClassType.StructMember &&
                TypeInfo.ClassType != ClassType.UnionMember && TypeInfo.ClassType != ClassType.EndOfStruct)
                throw new Exception($"Unexpected class {TypeInfo.ClassType}");
        }

        public void HandleInlineFakes(SymDumpProcessingConfiguration config)
        {
            if (config.AutoRenameFakes && (OriginalWasFake))
            {
                string cTypeBase = TypeInfo.AsCode("");
                string cType = "unknown";
                string cSubType = "unknown";
                if (cTypeBase.Contains(" "))
                {
                    string[] cTypeArray = cTypeBase.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
                    cType = cTypeArray[0];
                    if (cTypeArray.Length > 1)
                    {
                        cSubType = cTypeArray[1];
                    }
                }
                string parentType = "unknown";
                if (TypeInfo.ClassType == ClassType.StructMember)
                {
                    parentType = "struct";
                }
                if (TypeInfo.ClassType == ClassType.UnionMember)
                {
                    parentType = "union";
                }
                List<string> cSubTypes = new List<string>();
                cSubTypes.Add(cSubType);
                // get the original fake name of the parent
                string originalCSubType = cSubType;
                int ulPos = cSubType.LastIndexOf("_");
                if (ulPos > 0)
                {
                    originalCSubType = originalCSubType.Substring(ulPos);
                    cSubTypes.Add(originalCSubType);
                }
                bool foundMatch = false;
                if (cType == "struct")
                {
                    for (int i = 0; i < cSubTypes.Count; i++)
                    {
                        if (config.FakeStructs.ContainsKey(cSubTypes[i]))
                        {
                            _InlineStructDef = config.FakeStructs[cSubTypes[i]];
                            SymFile.OutputDebugMessage(string.Format("Matched fake {0} '{1} {2}' in {3} '{4}' with existing fake '{5}'",
                                cType, cSubType, name, parentType, ParentName, cSubTypes[i]), config);
                            config.FakeStructs[cSubTypes[i]].HasBeenInlined = true;
                            foundMatch = true;
                            break;
                        }
                    }
                }
                if (cType == "union")
                {
                    for (int i = 0; i < cSubTypes.Count; i++)
                    {
                        if (config.FakeUnions.ContainsKey(cSubTypes[i]))
                        {
                            _InlineUnionDef = config.FakeUnions[cSubTypes[i]];
                            SymFile.OutputDebugMessage(string.Format("Matched fake {0} '{1} {2}' in {3} '{4}' with existing fake '{5}'",
                                cType, cSubType, name, parentType, ParentName, cSubTypes[i]), config);
                            config.FakeUnions[cSubTypes[i]].HasBeenInlined = true;
                            foundMatch = true;
                            break;
                        }
                    }
                }
                if (!foundMatch)
                {
                    SymFile.OutputDebugMessage(string.Format("Error: did not find fake {0} '{1}' when populating {1} '{2}'", cType, cSubTypes, parentType, ParentName), config);
                }
            }
        }

        public bool IsFake()
        {
            if (TypeInfo == null)
            {
                return false;
            }
            if (TypeInfo.Tag == null)
            {
                return false;
            }
            //Regex fakeRex = new Regex(@"^(\.|_)\d+fake$");
            Regex fakeRex = new Regex(@"(\.|_)\d+fake$");
            return fakeRex.IsMatch(TypeInfo.Tag);
            //return fakeRex.IsMatch(name);
        }

        public bool Equals(StructMember other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            bool result = true;
            if (!TypeInfo.Equals(other.TypeInfo))
            {
                SymFile.OutputDebugMessage(string.Format("StructMember objects don't have matching TypeInfo objects ('{0}' vs '{1}')", TypeInfo.ToString(), other.TypeInfo.ToString()), ProcessingConfig);
                result = false;
            }
            if (!TypedValue.Equals(other.TypedValue))
            {
                SymFile.OutputDebugMessage(string.Format("StructMember objects don't have matching TypedValue objects ('{0}' vs '{1}')", TypedValue.ToString(), other.TypedValue.ToString()), ProcessingConfig);
                result = false;
            }
            if ((!OriginalWasFake) || (!other.OriginalWasFake))
            {
                if (name != other.name)
                {
                    SymFile.OutputDebugMessage(string.Format("StructMember objects don't have matching names ('{0}' vs '{1}')", name, other.name), ProcessingConfig);
                    result = false;
                }
            }
            return result;
        }

        public override string ToString()
        {
            switch (TypeInfo.ClassType)
            {
                case ClassType.Bitfield:
                    return TypeInfo.AsCode(name) +
                           $" : {TypeInfo.Size}; // offset={TypedValue.Value / 8}.{TypedValue.Value % 8}";
                case ClassType.StructMember:
                case ClassType.UnionMember:
                    //return TypeInfo.AsCode(name) +
                    //      $"; // size={TypeInfo.Size}, offset={TypedValue.Value}";
                    return GetSignature();
                default:
                    throw new Exception($"Unexpected class {TypeInfo.ClassType}");
            }
        }

        public string GetClassTypeName()
        {
            switch (TypeInfo.ClassType)
            {
                case ClassType.Bitfield:
                    return "bitfield";
                case ClassType.StructMember:
                    return "struct_member";
                case ClassType.UnionMember:
                    return "union_member";
                default:
                    break;
            }
            throw new Exception($"Unexpected class {TypeInfo.ClassType}");
        }

        protected string GetSignature()
        {
            string result = "";
            bool handledAsInlineFake = false;
            if (_InlineFakes && OriginalWasFake)
            {
                if (_InlineStructDef != null)
                {
                    result = _InlineStructDef.GetAsInlineStruct(name, TypeInfo, TypedValue);
                    handledAsInlineFake = true;
                }
                if (!handledAsInlineFake)
                {
                    if (_InlineUnionDef != null)
                    {
                        result = _InlineUnionDef.GetAsInlineUnion(name, TypeInfo, TypedValue);
                        handledAsInlineFake = true;
                    }
                }
            }
            if (!handledAsInlineFake)
            {
                result = TypeInfo.AsCode(name) + $"; // size={TypeInfo.Size}, offset={TypedValue.Value}"; ;
            }

            return result;
        }

        public BenLincoln.DebugSymbols.SymDump.CStructMember ToBLDSCStructMember()
        {
            BenLincoln.DebugSymbols.SymDump.CStructMember result = new BenLincoln.DebugSymbols.SymDump.CStructMember();
            result.name = name;
            //result.c_type = TypeInfo.GetCType();
            //result.type_name = TypeInfo.GetName(_name);
            result.struct_member_signature = GetSignature();
            //BenLincoln.DebugSymbols.SymDump.CParameter[] pParsed = BenLincoln.DebugSymbols.SymDump.CElement.GetCParametersFromString(result.struct_member_signature, true);
            //result.c_type = pParsed[0].type_name;
            //result.type_name = pParsed[0].type_name;
            result.c_type = TypeInfo.AsCode("");
            result.type_name = result.c_type;
            result.parent_name = ParentName;
            result.hashcode = GetHashCode();
            result.class_type = GetClassTypeName();
            result.size = TypeInfo.Size;
            result.fake = OriginalWasFake;
            result.derived_types = TypeInfo.GetDerivedTypes(name).ToArray();
            result.sym_file_sequence_number = TypedValue.SYMFileSequenceNumber;
            // damn it, Psygnosis, or whoever decided to use 0 for "default" instead of including the actual size
            if (result.size == 0)
            {
                bool handled = false;
                switch(result.c_type)
                {
                    case "byte":
                    case "char":
                    case "unsigned char":
                        result.size = 1;
                        handled = true;
                        break;
                    case "short":
                    case "unsigned short":
                        result.size = 2;
                        handled = true;
                        break;
                    case "float":
                    case "int":
                    case "long":
                    case "unsigned int":
                        result.size = 4;
                        handled = true;
                        break;
                    case "double":
                    case "unsigned long":
                        result.size = 8;
                        handled = true;
                        break;
                    case "void":
                    case "NULL":
                        // not sure this is correct
                        result.size = 4;
                        handled = true;
                        break;
                }
                // catch pointers
                if (!handled)
                {
                    if (result.c_type.EndsWith("*"))
                    {
                        result.size = 4;
                        handled = true;
                    }
                }

                // catch function pointers
                if (!handled)
                {
                    if (result.c_type.EndsWith("(*)()"))
                    {
                        result.size = 4;
                        handled = true;
                    }
                    if (result.c_type.EndsWith("(**)()"))
                    {
                        result.size = 4;
                        handled = true;
                    }
                    if (result.c_type.EndsWith("(***)()"))
                    {
                        result.size = 4;
                        handled = true;
                    }
                }

                // catch zero-size (variable-length) arrays, which, who knows? I'm just going to leave them set to zero, because, whatever, C standards committee
                // and Psygnosis, and anyone else involved in this. Thanks for making things difficult.
                if (!handled)
                {
                    if ((result.c_type.EndsWith("[0]")) || (result.c_type.EndsWith("[0])")))
                    {
                        result.size = 0;
                        handled = true;
                    }
                }

                if (!handled)
                {
                    throw new Exception(String.Format("Unknown type: '{0}' - cannot set size", result.c_type));
                }
            }
            result.offset_in_memory = TypedValue.Value;

            return result;
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((StructMember) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                var hashCode = name != null ? name.GetHashCode() : 0;
                hashCode = (hashCode * 397) ^ (TypeInfo != null ? TypeInfo.GetHashCode() : 0);
                hashCode = (hashCode * 397) ^ (TypedValue != null ? TypedValue.GetHashCode() : 0);
                return hashCode;
            }
        }
    }
}
