﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using symdump.symfile.util;
using symdump.util;

namespace symdump.symfile
{
    public class DependentElementProcessingResult
    {
        public Dictionary<int, StructDef> StructCollection;
        public Dictionary<int, UnionDef> UnionCollection;
        public List<int> StructSequenceNumbers;
        public List<int> UnionSequenceNumbers;

        public DependentElementProcessingResult()
        {
            StructCollection = new Dictionary<int, StructDef>();
            UnionCollection = new Dictionary<int, UnionDef>();
            StructSequenceNumbers = new List<int>();
            UnionSequenceNumbers = new List<int>();
        }

        public static DependentElementProcessingResult GetDependentElements(Dictionary<string, StructDef> structs, Dictionary<string, UnionDef> unions, SymDumpProcessingConfiguration config)
        {
            DependentElementProcessingResult result = new DependentElementProcessingResult();
            foreach (string k in structs.Keys)
            {
                if (result.StructSequenceNumbers.Contains(structs[k].SYMFileSequenceNumber))
                {
                    throw new Exception(string.Format("List of struct SYM file sequence numbers already contains {0}", (structs[k].SYMFileSequenceNumber)));
                }
                result.StructSequenceNumbers.Add(structs[k].SYMFileSequenceNumber);
                result.StructCollection.Add(structs[k].SYMFileSequenceNumber, structs[k]);
            }
            result.StructSequenceNumbers.Sort();
            foreach (string k in unions.Keys)
            {
                if (result.UnionSequenceNumbers.Contains(unions[k].SYMFileSequenceNumber))
                {
                    throw new Exception(string.Format("List of union SYM file sequence numbers already contains {0}", (unions[k].SYMFileSequenceNumber)));
                }
                result.UnionSequenceNumbers.Add(unions[k].SYMFileSequenceNumber);
                result.UnionCollection.Add(unions[k].SYMFileSequenceNumber, unions[k]);
            }
            result.UnionSequenceNumbers.Sort();
            if (config.AutoRenameFakes)
            {
                Dictionary<string, string> structNameMap = new Dictionary<string, string>();
                Dictionary<string, string> unionNameMap = new Dictionary<string, string>();

                foreach (string k in structs.Keys)
                {
                    //Console.WriteLine(k);
                    //_structs[k].Members[0];
                }
                //for (int i = 0; i < result.structs.Count; i++)
                //{

                //}
            }

            return result;
        }
    }

    public class ObjectFileData
    {
        public long MinAddress;
        public long MaxAddress;
        public List<long> Addresses;
        public string FileName;

        public ObjectFileData()
        {
            MinAddress = 0xFFFFFFFF;
            MaxAddress = 0x00000000;
            Addresses = new List<long>();
            FileName = "";
        }

        public void UpdateAddressData(long newAddress)
        {
            if (!Addresses.Contains(newAddress))
            {
                Addresses.Add(newAddress);
            }
            if (newAddress < MinAddress)
            {
                MinAddress = newAddress;
            }
            if (newAddress > MaxAddress)
            {
                MaxAddress = newAddress;
            }
        }

        public void Dump(IndentedTextWriter writer)
        {
            writer.WriteLine(string.Format("// Object file info - file '{0}', min address = 0x{1:X8}, max address = 0x{2:X8}", FileName, MinAddress, MaxAddress));
        }

        public BenLincoln.DebugSymbols.SymDump.ObjectFileData ToBLDSObjectFileData()
        {
            BenLincoln.DebugSymbols.SymDump.ObjectFileData result = new BenLincoln.DebugSymbols.SymDump.ObjectFileData();
            result.file_name = FileName;
            Addresses.Sort();
            for (int i = 0; i < Addresses.Count; i++)
            {
                result.addresses.Add(Addresses[i]);
            }
            result.min_address = MinAddress;
            result.max_address = MaxAddress;
            if (MinAddress == 0xFFFFFFFF)
            {
                result.min_address = -1;
            }
            if (MaxAddress == 0x0)
            {
                result.max_address = -1;
            }
            return result;
        }
    }

    public struct SLDAddressLineNumberSet
    {
        public long Address;
        public long LineNumber;
    }

    public class SLDDataSet
    {
        public string SLDFile;
        public List<SLDAddressLineNumberSet> AddressesAndLineNumbers;

        public SLDDataSet()
        {
            SLDFile = "";
            AddressesAndLineNumbers = new List<SLDAddressLineNumberSet>();
        }

        public SLDDataSet Clone()
        {
            SLDDataSet result = new SLDDataSet();
            result.SLDFile = SLDFile;
            result.AddressesAndLineNumbers = new List<SLDAddressLineNumberSet>();
            for (int i = 0; i < AddressesAndLineNumbers.Count; i++)
            {
                SLDAddressLineNumberSet newSLDANS = new SLDAddressLineNumberSet();
                newSLDANS.Address = AddressesAndLineNumbers[i].Address;
                newSLDANS.LineNumber = AddressesAndLineNumbers[i].LineNumber;
                result.AddressesAndLineNumbers.Add(newSLDANS);
            }
            return result;
        }

        public void Dump(IndentedTextWriter writer)
        {
            writer.WriteLine(string.Format("// SLD info, file '{0}'", SLDFile));
            ++writer.Indent;
            foreach (SLDAddressLineNumberSet sldalns in AddressesAndLineNumbers)
            {
                writer.WriteLine(string.Format("// Address: 0x{0:X8}, line number: {1}", sldalns.Address, sldalns.LineNumber));
            }
            --writer.Indent;
        }
    }

    public class PsyQOverlay
    {
        public long OverlayID;
        public long Offset;
        public long Length;

        public Dictionary<string, EnumDef> Enums = new Dictionary<string, EnumDef>();
        public List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration> Externs = new List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration>();
        public List<Function> Functions = new List<Function>();

        public Dictionary<uint, List<Label>> Labels = new Dictionary<uint, List<Label>>();
        public Dictionary<string, StructDef> Structs = new Dictionary<string, StructDef>();

        public Dictionary<string, TypeInfo> Typedefs = new Dictionary<string, TypeInfo>();
        public Dictionary<string, UnionDef> Unions = new Dictionary<string, UnionDef>();

        public List<SLDDataSet> SLDEntries;

        public List<ObjectFileData> ObjectFileDataEntries;

        public PsyQOverlay()
        {
            OverlayID = -1;
            Offset = -1;
            Length = -1;

            Enums = new Dictionary<string, EnumDef>();
            Externs = new List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration>();
            Functions = new List<Function>();
            Labels = new Dictionary<uint, List<Label>>();
            Structs = new Dictionary<string, StructDef>();
            Typedefs = new Dictionary<string, TypeInfo>();
            Unions = new Dictionary<string, UnionDef>();

            SLDEntries = new List<SLDDataSet>();
            ObjectFileDataEntries = new List<ObjectFileData>();
        }
    }

    public class SymDumpProcessingConfiguration
    {
        public SymbolParsingTracker Tracker;
        public BenLincoln.DebugSymbols.SequenceIssuer Issuer;
        public List<StructDef> AllStructs;
        public List<UnionDef> AllUnions;
        public Dictionary<string, StructDef> FakeStructs;
        public Dictionary<string, UnionDef> FakeUnions;
        public bool RenameForCompatibility;
        public bool IgnoreDuplicateDefinitions;
        public bool AutoRenameFakes;
        public bool InlineFakes;
        public bool ShowDebugMessages;
        public bool ShowInfoMessages;
        public bool ShowWarningMessages;
        public bool ShowErrorMessages;

        public SymDumpProcessingConfiguration()
        {
            Tracker = new SymbolParsingTracker();
            Issuer = new BenLincoln.DebugSymbols.SequenceIssuer();
            AllStructs = new List<StructDef>();
            AllUnions = new List<UnionDef>();
            FakeStructs = new Dictionary<string, StructDef>();
            FakeUnions = new Dictionary<string, UnionDef>();
            RenameForCompatibility = false;
            IgnoreDuplicateDefinitions = false;
            AutoRenameFakes = false;
            InlineFakes = false;
            ShowDebugMessages = false;
            ShowInfoMessages = true;
            ShowWarningMessages = true;
            ShowErrorMessages = true;
        }
    }

    public class SymbolParsingTracker
    {
        public List<SLDDataSet> SLDEntries;
        protected SLDDataSet _CurrentSLDEntry;
        public List<ObjectFileData> ObjectFileDataEntries;
        
        protected bool _FirstSLDEntryAlreadySet;
        protected bool _FirstObjectFileEntryAlreadySet;
        public string CurrentSLDFile;
        public long CurrentSLDAddress;
        public long CurrentSLDLineNumber;
        public long CurrentOverlayID;
        public ObjectFileData CurrentObjectFileData;

        public Dictionary<long, PsyQOverlay> Overlays;

        public SymbolParsingTracker()
        {
            CurrentSLDFile = "";
            CurrentSLDAddress = -1;
            CurrentSLDLineNumber = -1;
            CurrentOverlayID = -1;
            SLDEntries = new List<SLDDataSet>();
            _CurrentSLDEntry = new SLDDataSet();
            _FirstSLDEntryAlreadySet = false;
            _FirstObjectFileEntryAlreadySet = false;
            ObjectFileDataEntries = new List<ObjectFileData>();
            CurrentObjectFileData = new ObjectFileData();

            Overlays = new Dictionary<long, PsyQOverlay>();
        }

        public void SetSLDFile(string fileName, uint address, uint lineNumber)
        {
            HandleSLDEnd();
            _CurrentSLDEntry.SLDFile = fileName;
            CurrentSLDFile = fileName;
            _CurrentSLDEntry.AddressesAndLineNumbers = new List<SLDAddressLineNumberSet>();
            SetSLDLineNumber(address, lineNumber);
            CurrentSLDAddress = address;
            CurrentSLDLineNumber = lineNumber;
            _FirstSLDEntryAlreadySet = true;
        }

        public void HandleObjectFileFileName(string fileName)
        {
            bool startNewObjectFileEntry = false;
            if (CurrentObjectFileData.FileName == fileName)
            {
                if (CurrentOverlayID == -1)
                {
                    ObjectFileDataEntries.Add(CurrentObjectFileData);
                }
                else
                {
                    Overlays[CurrentOverlayID].ObjectFileDataEntries.Add(CurrentObjectFileData);
                }
                CurrentObjectFileData = new ObjectFileData();
                _FirstObjectFileEntryAlreadySet = true;
            }
            else
            {
                startNewObjectFileEntry = true;
            }
            if (startNewObjectFileEntry)
            {
                CurrentObjectFileData = new ObjectFileData();
                CurrentObjectFileData.FileName = fileName;
            }
        }

        public void SetSLDLineNumber(uint address, uint lineNumber)
        {
            SLDAddressLineNumberSet newSLDANS = new SLDAddressLineNumberSet();
            newSLDANS.Address = address;
            newSLDANS.LineNumber = lineNumber;
            _CurrentSLDEntry.AddressesAndLineNumbers.Add(newSLDANS);
        }

        public void IncrementSLDLineNumber(uint address, long amount)
        {
            CurrentSLDLineNumber += amount;
            SLDAddressLineNumberSet newSLDANS = new SLDAddressLineNumberSet();
            newSLDANS.Address = address;
            newSLDANS.LineNumber = CurrentSLDLineNumber;
            _CurrentSLDEntry.AddressesAndLineNumbers.Add(newSLDANS);
        }

        public void HandleSLDEnd()
        {
            if (_FirstSLDEntryAlreadySet)
            {
                if ((_CurrentSLDEntry.SLDFile != "") || (_CurrentSLDEntry.AddressesAndLineNumbers.Count > 0))
                {
                    SLDDataSet addDS = _CurrentSLDEntry.Clone();
                    if (CurrentOverlayID == -1)
                    {
                        SLDEntries.Add(addDS);
                    }
                    else
                    {
                        Overlays[CurrentOverlayID].SLDEntries.Add(addDS);
                    }
                }
            }
            _CurrentSLDEntry = new SLDDataSet();
            CurrentSLDFile = "";
            CurrentSLDAddress = -1;
            CurrentSLDLineNumber = -1;
        }

        public void AddOverlay(long id, long offset, long length)
        {
            PsyQOverlay newOverlay = new PsyQOverlay();
            newOverlay.OverlayID = id;
            newOverlay.Offset = offset;
            newOverlay.Length = length;
            Overlays.Add(id, newOverlay);
        }
    }

    public class SymFile
    {
        private Dictionary<string, EnumDef> _enums = new Dictionary<string, EnumDef>();
        //private readonly SortedSet<string> _externs = new SortedSet<string>();
        private List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration> _externs = new List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration>();
        public List<Function> Functions = new List<Function>();
        private Dictionary<string, string> _funcTypes = new Dictionary<string, string>();
        internal Dictionary<uint, List<Label>> Labels = new Dictionary<uint, List<Label>>();
        private Dictionary<string, StructDef> _structs = new Dictionary<string, StructDef>();
        private byte _targetUnit;
        private Dictionary<string, TypeInfo> _typedefs = new Dictionary<string, TypeInfo>();
        private Dictionary<string, UnionDef> _unions = new Dictionary<string, UnionDef>();
        private byte _version;
        private string _mxInfo;
        List<BenLincoln.ExecutableFiles.MemorySegment> MemorySegments { get; set; }
        public uint ProgramCounter { get; set; }
        public uint GlobalPointer { get; set; }
        public uint StackPointerBase { get; set; }
        public uint StackPointerSizeOrOffset { get; set; }
        protected Dictionary<long, PsyQOverlay> _overlays;
        protected List<SLDDataSet> _sldData;
        protected List<ObjectFileData> _objectFileData;

        public SymFile(BinaryReader stream, SymDumpProcessingConfiguration config)
        {
            MemorySegments = new List<BenLincoln.ExecutableFiles.MemorySegment> { };
            stream.BaseStream.Seek(0, SeekOrigin.Begin);

            stream.Skip(3);
            _version = stream.ReadByte();
            _targetUnit = stream.ReadByte();

            stream.Skip(3);

            while (stream.BaseStream.Position < stream.BaseStream.Length)
            {
                config.Tracker = DumpEntry(stream, config);
            }

            // post-process structs and unions to handle inline fakes
            for (int structNum = 0; structNum < config.AllStructs.Count; structNum++)
            {
                _structs = ProcessStruct(config.AllStructs[structNum], _structs, config);
            }
            for (int unionNum = 0; unionNum < config.AllUnions.Count; unionNum++)
            {
                _unions = ProcessUnion(config.AllUnions[unionNum], _unions, config);
            }
            // now that everything has been renamed, do a second loop for the actual inlining
            if (config.InlineFakes)
            {
                for (int structNum = 0; structNum < config.AllStructs.Count; structNum++)
                {
                    for (int memNum = 0; memNum < config.AllStructs[structNum].Members.Count; memNum++)
                    {
                        config.AllStructs[structNum].Members[memNum].HandleInlineFakes(config);
                    }
                }
                for (int unionNum = 0; unionNum < config.AllUnions.Count; unionNum++)
                {
                    for (int memNum = 0; memNum < config.AllUnions[unionNum].Members.Count; memNum++)
                    {
                        config.AllUnions[unionNum].Members[memNum].HandleInlineFakes(config);
                    }
                }
            }

            _overlays = config.Tracker.Overlays;
            _sldData = config.Tracker.SLDEntries;
            _objectFileData = config.Tracker.ObjectFileDataEntries;
        }

        public void GetDataFromPSXEXE(string psxEXEPath)
        {
            BenLincoln.ExecutableFiles.BinaryFile inputFile = new BenLincoln.ExecutableFiles.BinaryFile();
            BenLincoln.ExecutableFiles.PSXEXE inputFilePSXEXE = new BenLincoln.ExecutableFiles.PSXEXE();
            string exceptionMessagePSXEXE = "";

            try
            {
                FileStream inputStream = new FileStream(psxEXEPath, FileMode.Open, FileAccess.Read);
                BinaryReader inputReader = new BinaryReader(inputStream);

                try
                {
                    inputFilePSXEXE = new BenLincoln.ExecutableFiles.PSXEXE();
                    inputFilePSXEXE.ReadFile(inputReader);
                    
                    int segNum = 1;

                    // text segment
                    BenLincoln.ExecutableFiles.MemorySegment msText = new BenLincoln.ExecutableFiles.MemorySegment();
                    msText.SegmentNumber = segNum;
                    msText.Name = "text";
                    msText.Offset = inputFilePSXEXE.PSXExec.TextSegmentAddress;
                    msText.Length = inputFilePSXEXE.PSXExec.TextSegmentSize;
                    MemorySegments.Add(msText);
                    segNum++;

                    ProgramCounter = inputFilePSXEXE.PSXExec.ProgramCounter;
                    GlobalPointer = inputFilePSXEXE.PSXExec.GlobalPointer;
                    StackPointerBase = inputFilePSXEXE.PSXExec.StackSegmentAddress;
                    StackPointerSizeOrOffset = inputFilePSXEXE.PSXExec.StackSegmentSize;

                    // segments that usually don't exist
                    if (inputFilePSXEXE.PSXExec.DataSegmentAddress != 0)
                    {
                        BenLincoln.ExecutableFiles.MemorySegment msData = new BenLincoln.ExecutableFiles.MemorySegment();
                        msData.SegmentNumber = segNum;
                        msData.Name = "data";
                        msData.Offset = inputFilePSXEXE.PSXExec.DataSegmentAddress;
                        msData.Length = inputFilePSXEXE.PSXExec.DataSegmentSize;
                        MemorySegments.Add(msData);
                        segNum++;
                    }

                    if (inputFilePSXEXE.PSXExec.BSSSegmentAddress != 0)
                    {
                        BenLincoln.ExecutableFiles.MemorySegment msBSS = new BenLincoln.ExecutableFiles.MemorySegment();
                        msBSS.SegmentNumber = segNum;
                        msBSS.Name = "bss";
                        msBSS.Offset = inputFilePSXEXE.PSXExec.BSSSegmentAddress;
                        msBSS.Length = inputFilePSXEXE.PSXExec.BSSSegmentSize;
                        MemorySegments.Add(msBSS);
                        segNum++;
                    }

                    // RAM
                    BenLincoln.ExecutableFiles.MemorySegment msStack = new BenLincoln.ExecutableFiles.MemorySegment();
                    msStack.SegmentNumber = segNum;
                    msStack.Name = "ram";
                    //msStack.Offset = 0x801F8000;
                    msStack.Offset = inputFilePSXEXE.PSXExec.TextSegmentAddress + inputFilePSXEXE.PSXExec.TextSegmentSize;
                    if (msStack.Offset > 0x80200000)
                    {
                        throw new Exception(String.Format("Unsupported configuration: text segment address + length (0x{0:X})is greater than 0x80200000", msStack.Offset));
                    }
                    uint ramEndAddress = 0x80200000;
                    // this is a hack, sorry
                    if (msStack.Offset < 0x00200000)
                    {
                        ramEndAddress = 0x00200000;
                    }
                    //else
                    //{
                    //    throw new Exception(String.Format("Unsupported configuration: couldn't calculate valid RAM end address for stack offset 0x{0:X8}", msStack.Offset));
                    //}
                    msStack.Length = ramEndAddress - msStack.Offset;
                    MemorySegments.Add(msStack);
                    segNum++;

                    // RAM

                }
                catch (Exception fe)
                {
                    exceptionMessagePSXEXE = fe.Message;
                }

            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Could not read the input file '{0}': {1}", psxEXEPath, ex.Message));
            }
        }

        public void Dump(TextWriter output)
        {
            var writer = new IndentedTextWriter(output);
            writer.WriteLine($"Version = {_version}, targetUnit = {_targetUnit}");

            DumpCollection(writer, _enums, _unions, _structs, _typedefs, Labels, _externs, Functions, _sldData, _objectFileData);

            foreach(int overlayID in _overlays.Keys)
            {
                PsyQOverlay ol = _overlays[overlayID];
                writer.WriteLine();
                writer.WriteLine(string.Format("// begin overlay 0x{0:X2}, offset 0x{1:X8}, length 0x{2:X8} (decimal: {2})", ol.OverlayID, ol.Offset, ol.Length));

                DumpCollection(writer, ol.Enums, ol.Unions, ol.Structs, ol.Typedefs, ol.Labels, ol.Externs, ol.Functions, ol.SLDEntries, ol.ObjectFileDataEntries);

                writer.WriteLine(string.Format("// end overlay 0x{0:X2}", ol.OverlayID));
                writer.WriteLine();
            }
        }

        public static void DumpCollection(IndentedTextWriter writer, Dictionary<string, EnumDef> enums, Dictionary<string, UnionDef> unions, Dictionary<string, StructDef> structs
            , Dictionary<string, TypeInfo> typedefs, Dictionary<uint, List<Label>> labels, List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration> externs, List<Function> functions,
            List<SLDDataSet> sldData, List<ObjectFileData> objectFileData)
        {
            writer.WriteLine();
            writer.WriteLine($"// {enums.Count} enums");
            foreach (var e in enums.Values)
            {
                e.Dump(writer);
            }

            writer.WriteLine();
            writer.WriteLine($"// {unions.Count} unions");
            foreach (var e in unions.Values)
            {
                e.Dump(writer);
            }

            writer.WriteLine();
            writer.WriteLine($"// {structs.Count} structs");
            foreach (var e in structs.Values)
            {
                e.Dump(writer);
            }

            writer.WriteLine();
            writer.WriteLine($"// {typedefs.Count} typedefs");
            foreach (var t in typedefs)
            {
                writer.WriteLine($"typedef {t.Value.AsCode(t.Key)};");
            }

            writer.WriteLine();
            writer.WriteLine($"// {labels.Count} labels");
            foreach (var l in labels)
            {
                foreach (var l2 in l.Value)
                {
                    writer.WriteLine(l2);
                }
            }

            writer.WriteLine();
            writer.WriteLine($"// {externs.Count} external declarations");
            foreach (var e in externs)
            {
                writer.WriteLine(e);
            }

            writer.WriteLine();
            writer.WriteLine($"// {functions.Count} functions");
            foreach (var f in functions)
            {
                f.Dump(writer);
            }

            writer.WriteLine();
            writer.WriteLine($"// {sldData.Count} sets of SLD information");
            foreach (SLDDataSet s in sldData)
            {
                s.Dump(writer);
            }

            writer.WriteLine();
            writer.WriteLine($"// {objectFileData.Count} sets of object file information");
            foreach (ObjectFileData ofd in objectFileData)
            {
                ofd.Dump(writer);
            }
        }

        public BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection ToBLDSSymDumpSymbolCollection(string defaultSourceFileNameBase, SymDumpProcessingConfiguration config)
        {
            BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection result = new BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection();

            //string defaultSourceFileHeader = String.Format("{0}.H", defaultSourceFileNameBase);
            //string defaultSourceFileCode = String.Format("{0}.C", defaultSourceFileNameBase);

            string defaultSourceFileHeader = "";
            string defaultSourceFileCode = "";

            result.version = _version;
            result.target_unit = _targetUnit;
            result.program_counter = ProgramCounter;
            result.global_pointer = GlobalPointer;
            result.stack_pointer_base = StackPointerBase;
            result.stack_pointer_size_or_offset = StackPointerSizeOrOffset;

            // directly from EXE
            result.memory_segments = new List<BenLincoln.DebugSymbols.SymDump.MemorySegment> { };
            if (MemorySegments.Count > 0)
            {
                for (int i = 0; i < MemorySegments.Count; i++)
                {
                    BenLincoln.DebugSymbols.SymDump.MemorySegment ms = new BenLincoln.DebugSymbols.SymDump.MemorySegment();
                    ms.name = MemorySegments[i].Name;
                    ms.offset = MemorySegments[i].Offset;
                    ms.length = MemorySegments[i].Length;
                    ms.segment_number = MemorySegments[i].SegmentNumber;
                    result.memory_segments.Add(ms);
                }
            }

            DependentElementProcessingResult coreDependentElements = DependentElementProcessingResult.GetDependentElements(_structs, _unions, config);

            result.structs = GetStructs(coreDependentElements.StructCollection, coreDependentElements.StructSequenceNumbers, -1, config);
            result.unions = GetUnions(coreDependentElements.UnionCollection, coreDependentElements.UnionSequenceNumbers, -1, config);
            result.external_declarations = GetExternals(_externs, -1);
            result.enums = GetEnums(_enums, defaultSourceFileHeader, -1);
            result.typedefs = GetTypeDefs(_typedefs, defaultSourceFileHeader, -1);
            result.labels = GetLabels(Labels, -1);
            result.functions = GetFunctionList(Functions, defaultSourceFileCode, -1);
            result.sld_data = GetSLDData(_sldData, -1);
            result.object_file_data = new List<BenLincoln.DebugSymbols.SymDump.ObjectFileData>();
            for (int i = 0; i < _objectFileData.Count; i++)
            {
                BenLincoln.DebugSymbols.SymDump.ObjectFileData ofd = _objectFileData[i].ToBLDSObjectFileData();
                result.object_file_data.Add(ofd);
            }

            result.overlays = new List<BenLincoln.DebugSymbols.SymDump.PsyQOverlayDescription>();
            foreach (int overlayID in _overlays.Keys)
            {
                PsyQOverlay o = _overlays[overlayID];
                BenLincoln.DebugSymbols.SymDump.PsyQOverlayDescription newOverlay = new BenLincoln.DebugSymbols.SymDump.PsyQOverlayDescription();
                newOverlay.overlay_id = o.OverlayID;
                newOverlay.offset = o.Offset;
                newOverlay.length = o.Length;

                DependentElementProcessingResult overlayDependentElements = DependentElementProcessingResult.GetDependentElements(o.Structs, o.Unions, config);

                newOverlay.structs = GetStructs(overlayDependentElements.StructCollection, overlayDependentElements.StructSequenceNumbers, overlayID, config);
                newOverlay.unions = GetUnions(overlayDependentElements.UnionCollection, overlayDependentElements.UnionSequenceNumbers, overlayID, config);
                newOverlay.external_declarations = GetExternals(o.Externs, overlayID);
                newOverlay.enums = GetEnums(o.Enums, defaultSourceFileHeader, overlayID);
                newOverlay.typedefs = GetTypeDefs(o.Typedefs, defaultSourceFileHeader, overlayID);
                newOverlay.labels = GetLabels(o.Labels, overlayID);
                newOverlay.functions = GetFunctionList(o.Functions, defaultSourceFileCode, overlayID);
                newOverlay.sld_data = GetSLDData(o.SLDEntries, overlayID);

                result.overlays.Add(newOverlay);
            }
 

            return result;
        }

        protected static List<BenLincoln.DebugSymbols.SymDump.CStruct> GetStructs(Dictionary<int, StructDef> structCollection, List<int> structSequenceNumbers, int overlayID, 
            SymDumpProcessingConfiguration config)
        {
            List<BenLincoln.DebugSymbols.SymDump.CStruct> result = new List<BenLincoln.DebugSymbols.SymDump.CStruct>();
            for (int i = 0; i < structSequenceNumbers.Count; i++)
            {
                BenLincoln.DebugSymbols.SymDump.CStruct newStruct = structCollection[structSequenceNumbers[i]].ToBLDSCStruct(overlayID);
                newStruct.overlay_id = overlayID;
                if ((!config.InlineFakes) || (!structCollection[structSequenceNumbers[i]].HasBeenInlined))
                {
                    result.Add(newStruct);
                }
            }
            return result;
        }

        protected static List<BenLincoln.DebugSymbols.SymDump.CUnion> GetUnions(Dictionary<int, UnionDef> unionCollection, List<int> unionSequenceNumbers, int overlayID,
            SymDumpProcessingConfiguration config)
        {
            List<BenLincoln.DebugSymbols.SymDump.CUnion> result = new List<BenLincoln.DebugSymbols.SymDump.CUnion>();
            for (int i = 0; i < unionSequenceNumbers.Count; i++)
            {
                BenLincoln.DebugSymbols.SymDump.CUnion newUnion = unionCollection[unionSequenceNumbers[i]].ToBLDSCUnion(overlayID);
                newUnion.overlay_id = overlayID;
                if ((!config.InlineFakes) || (!unionCollection[unionSequenceNumbers[i]].HasBeenInlined))
                {
                    result.Add(newUnion);
                }
            }
            return result;
        }

        protected static List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration> GetExternals(List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration> externs, int overlayID)
        {
            List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration> result = new List<BenLincoln.DebugSymbols.SymDump.CExternalDeclaration>();
            foreach (BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x in externs)
            {
                x.overlay_id = overlayID;
                result.Add(x);
            }
            return result;
        }

        protected static List<BenLincoln.DebugSymbols.SymDump.CEnum> GetEnums(Dictionary<string, EnumDef> enums, string defaultSourceFile, int overlayID)
        {
            List<BenLincoln.DebugSymbols.SymDump.CEnum> result = new List<BenLincoln.DebugSymbols.SymDump.CEnum>();
            foreach (string k in enums.Keys)
            {
                EnumDef oe = enums[k];
                BenLincoln.DebugSymbols.SymDump.CEnum e = oe.ToBLDSEnumDef(defaultSourceFile);
                e.overlay_id = overlayID;
                result.Add(e);
            }

            return result;
        }

        protected static List<BenLincoln.DebugSymbols.SymDump.CTypeDef> GetTypeDefs(Dictionary<string, TypeInfo> typedefs, string defaultSourceFile, int overlayID)
        {
            List<BenLincoln.DebugSymbols.SymDump.CTypeDef> result = new List<BenLincoln.DebugSymbols.SymDump.CTypeDef>();

            foreach (string k in typedefs.Keys)
            {
                BenLincoln.DebugSymbols.SymDump.CTypeDef t = typedefs[k].ToBLDSCTypeDef(k, defaultSourceFile);
                // redefine pointer typedefs correctly
                if (t.name.StartsWith("*"))
                {
                    t.name = t.name.Substring(1);
                    t.c_type = String.Format("{0} *", t.c_type);
                }
                // redefine function pointer typedefs correctly (I think?)
                if (t.name.StartsWith("(*"))
                {
                    t.name = t.name.Substring(2);
                    int firstClosingParenPos = t.name.IndexOf(")");
                    t.name = t.name.Substring(0, firstClosingParenPos);
                    t.c_type = "void *";
                }
                t.overlay_id = overlayID;
                result.Add(t);
            }

            return result;
        }

        protected static List<BenLincoln.DebugSymbols.SymDump.CLabel> GetLabels(Dictionary<uint, List<Label>> labels, int overlayID)
        {
            List<BenLincoln.DebugSymbols.SymDump.CLabel> result = new List<BenLincoln.DebugSymbols.SymDump.CLabel>();
            foreach (uint k in labels.Keys)
            {
                List<Label> labelList = labels[k];
                foreach (Label l in labelList)
                {
                    BenLincoln.DebugSymbols.SymDump.CLabel newLabel = l.ToBLDSCLabel();
                    newLabel.label_type = l.LabelType;
                    newLabel.overlay_id = overlayID;
                    result.Add(newLabel);
                }
            }
            return result;
        }

        protected static List<BenLincoln.DebugSymbols.SymDump.PsyQSLDDataSet> GetSLDData(List<SLDDataSet> sldData, int overlayID)
        {
            List<BenLincoln.DebugSymbols.SymDump.PsyQSLDDataSet> result = new List<BenLincoln.DebugSymbols.SymDump.PsyQSLDDataSet>();

            foreach (SLDDataSet ds in sldData)
            {
                BenLincoln.DebugSymbols.SymDump.PsyQSLDDataSet newDS = new BenLincoln.DebugSymbols.SymDump.PsyQSLDDataSet();
                newDS.source_file = ds.SLDFile;
                newDS.data = new List<BenLincoln.DebugSymbols.SymDump.AddressAndLineNumberTuple>();
                foreach (SLDAddressLineNumberSet sldalns in ds.AddressesAndLineNumbers)
                {
                    BenLincoln.DebugSymbols.SymDump.AddressAndLineNumberTuple newSLDALNS = new BenLincoln.DebugSymbols.SymDump.AddressAndLineNumberTuple();
                    newSLDALNS.address = sldalns.Address;
                    newSLDALNS.line_number = sldalns.LineNumber;
                    newDS.data.Add(newSLDALNS);
                }
                newDS.overlay_id = overlayID;
                result.Add(newDS);
            }
            return result;
        }

        protected static List<BenLincoln.DebugSymbols.SymDump.CFunction> GetFunctionList(List<Function> functions, string defaultSourceFile, int overlayID)
        {
            List<BenLincoln.DebugSymbols.SymDump.CFunction> result = new List<BenLincoln.DebugSymbols.SymDump.CFunction>();
            foreach (Function f in functions)
            {
                BenLincoln.DebugSymbols.SymDump.CFunction func = f.ToBLDSCFunction(defaultSourceFile);
                func.overlay_id = overlayID;
                result.Add(func);
            }
            return result;
        }

        public string DumpJSON(string defaultSourceFileNameBase, SymDumpProcessingConfiguration config)
        {
            BenLincoln.DebugSymbols.SymDump.SymDumpSymbolCollection resultDef = ToBLDSSymDumpSymbolCollection(defaultSourceFileNameBase, config);
            BenLincoln.DebugSymbols.Utilities.HandleWarningForOverlays(resultDef);

            return resultDef.ToJSON();
        }

        public static void OutputMessage(string messageType, string message)
        {
            Console.WriteLine(string.Format("({0}) {1}", messageType, message));
        }

        public static void OutputDebugMessage(string message, SymDumpProcessingConfiguration config)
        {
            if (config.ShowDebugMessages)
            {
                OutputMessage("Debug", message);
            }
        }

        public static void OutputInfoMessage(string message, SymDumpProcessingConfiguration config)
        {
            if (config.ShowInfoMessages)
            {
                OutputMessage("Info", message);
            }
        }

        public static void OutputWarningMessage(string message, SymDumpProcessingConfiguration config)
        {
            if (config.ShowWarningMessages)
            {
                OutputMessage("Warning", message);
            }
        }

        public static void OutputErrorMessage(string message, SymDumpProcessingConfiguration config)
        {
            if (config.ShowErrorMessages)
            {
                OutputMessage("Error", message);
            }
        }

        private SymbolParsingTracker DumpEntry(BinaryReader stream, SymDumpProcessingConfiguration config)
        {
            long streamPosition = stream.BaseStream.Position;
            var typedValue = new TypedValue(stream, config);
            int tvt = typedValue.Type & 0x7f;
            OutputDebugMessage(string.Format("original entry type 0x{0:X2}, handled as type {1} by SymDump code, with value 0x{2:X4} @0x{3:X8}", 
                typedValue.Type, tvt, typedValue.Value, streamPosition), config);

            if (typedValue.Type == 8)
            {
                byte mxInfo = stream.ReadByte();
                string mxInfoString = $"${typedValue.Value:X} MX-info {mxInfo:X}";
                _mxInfo = mxInfoString;
                OutputDebugMessage(string.Format("Entry type {0}: MX-Info - {1:X}, {2:X}", tvt, typedValue.Value, mxInfo), config);
                return config.Tracker;
            }

            if (typedValue.IsLabel)
            {
                try
                {
                    var lbl = new Label(typedValue, stream);
                    lbl.SLDSourceFileName = config.Tracker.CurrentSLDFile;
                    lbl.ObjectFileName = config.Tracker.CurrentObjectFileData.FileName;
                    // type 1 labels are not good data for this
                    if (typedValue.Type > 1)
                    {
                        config.Tracker.CurrentObjectFileData.UpdateAddressData(lbl.Offset);
                    }
                    OutputDebugMessage(string.Format("Entry type {0}: Label (type 0x{1:X2}) at address 0x{2:X8}, '{3}'", tvt, typedValue.Type, lbl.Offset, lbl.Name), config);

                    if (config.Tracker.CurrentOverlayID == -1)
                    {
                        if (!Labels.ContainsKey(lbl.Offset))
                        {
                            Labels.Add(lbl.Offset, new List<Label>());
                        }
                        Labels[lbl.Offset].Add(lbl);
                    }
                    else
                    {
                        if (!config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Labels.ContainsKey(lbl.Offset))
                        {
                            config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Labels.Add(lbl.Offset, new List<Label>());
                        }
                        config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Labels[lbl.Offset].Add(lbl);
                    }
                }
                catch (Exception e)
                {
                    Console.WriteLine(string.Format("Error: Exception thrown while reading label data: {0}", e.Message));
                }
                return config.Tracker;
            }

            switch (tvt)
            {
                case 0:
                    OutputDebugMessage(string.Format("Entry type: {0}: Inc SLD linenum at address 0x{1:X8}", tvt, typedValue.Value), config);
                    config.Tracker.IncrementSLDLineNumber((uint)typedValue.Value, 1);
                    break;
                case 2:
                    byte lnIncByte = stream.ReadByte();
                    OutputDebugMessage(string.Format("Entry type: {0}: Inc SLD linenum by byte 0x{1:X} (decimal: {1}) at address 0x{2:X8}", tvt, lnIncByte, typedValue.Value), config);
                    config.Tracker.IncrementSLDLineNumber((uint)typedValue.Value, lnIncByte);
                    break;
                case 4:
                    ushort lnIncUShort = stream.ReadUInt16();
                    OutputDebugMessage(string.Format("Entry type: {0}: Inc SLD linenum by word 0x{1:X} (decimal: {1}) at address 0x{2:X8}", tvt, lnIncUShort, typedValue.Value), config);
                    config.Tracker.IncrementSLDLineNumber((uint)typedValue.Value, lnIncUShort);
                    break;
                case 6:
                    uint lnNewNumber = stream.ReadUInt32();
                    OutputDebugMessage(string.Format("Entry type: {0}: Set SLD linenum to 0x{1:X} (decimal: {1}) at address 0x{2:X8}", tvt, lnNewNumber, typedValue.Value), config);
                    config.Tracker.SetSLDLineNumber((uint)typedValue.Value, lnNewNumber);
                    break;
                case 8:
                    uint lnLineNumber = stream.ReadUInt32();
                    string lnIncUInt32FileName = FileUtil.ReadPascalString(stream);
                    OutputDebugMessage(string.Format("Entry type: {0}: Set SLD to line 0x{1:X} (decimal: {1}) of file '{2}' at address 0x{3:X8}", tvt, lnLineNumber, lnIncUInt32FileName, typedValue.Value), config);
                    config.Tracker.SetSLDFile(lnIncUInt32FileName, (uint)typedValue.Value, lnLineNumber);
                    break;
                case 10:    // 0x0A
                    OutputDebugMessage(string.Format("Entry type: {0}: End SLD info at address 0x{1:X8}", tvt, typedValue.Value), config);
                    config.Tracker.HandleSLDEnd();
                    break;
                case 12:    // 0x0C
                    OutputDebugMessage(string.Format("Entry type {0}: Function start at address 0x{1:X8}", tvt, typedValue.Value), config);
                    DumpType12(stream, typedValue.Value, config);
                    break;
                case 14:    // 0x0E
                    uint funcEndLine = stream.ReadUInt32();
                    Console.WriteLine($"Warning: stub handler for type number: {tvt}: Function end, address 0x{typedValue.Value:X8} line 0x{funcEndLine:X} (decimal: {funcEndLine})");
                    break;
                case 16:    // 0x10
                    uint blockStartLine = stream.ReadUInt32();
                    Console.WriteLine($"Warning: stub handler for type number: {tvt}: Block start, address 0x{typedValue.Value:X8}, line 0x{blockStartLine:X} (decimal: {blockStartLine})");
                    break;
                case 18:    // 0x12
                    uint blockEndLine = stream.ReadUInt32();
                    Console.WriteLine($"Warning: stub handler for type number: {tvt}: Block end, address 0x{typedValue.Value:X8}, line 0x{blockEndLine:X} (decimal: {blockEndLine})");
                    break;
                case 20:    // 0x14
                    OutputDebugMessage(string.Format("Entry type {0}: 0x{1:X8} Def", tvt, typedValue.Value), config);
                    DumpType20(stream, typedValue.Value, config);
                    break;
                case 22:    // 0x16
                    OutputDebugMessage(string.Format("Entry type {0}: 0x{1:X8} Def 2", tvt, typedValue.Value), config);
                    DumpType22(stream, typedValue.Value, config);
                    break;
                case 24:    // 0x18
                    uint overlayLength = stream.ReadUInt32();
                    uint overlayID = stream.ReadUInt32();
                    OutputDebugMessage(string.Format("Entry type {0}: overlay ID 0x{1:X} (decimal: {1}), base address 0x{2:X8}, length 0x{3:X} (decimal: {3})", tvt, overlayID, typedValue.Value, overlayLength), config);
                    config.Tracker.AddOverlay(overlayID, BenLincoln.DebugSymbols.Utilities.CastIntToUInt(typedValue.Value), overlayLength);
                    break;
                case 26:    // 0x1A
                    OutputDebugMessage(string.Format("Entry type {0}: set current overlay to 0x{1:X}", tvt, typedValue.Value), config);
                    config.Tracker.CurrentOverlayID = typedValue.Value;
                    if (typedValue.Value == 0)
                    {
                        // 0 in the SYM data means "no overlay", whereas I'm using -1
                        config.Tracker.CurrentOverlayID = -1;
                    }
                    break;
                case 28:    // 0x1C
                    ushort f2FP = stream.ReadUInt16();
                    uint f2Size = stream.ReadUInt32();
                    uint f2RetReg = stream.ReadUInt16();
                    uint f2Mask = stream.ReadUInt32();
                    uint f2MaskOffset = stream.ReadUInt32();
                    uint f2FMask = stream.ReadUInt32();
                    uint f2FMaskOffset = stream.ReadUInt32();
                    uint f2Line = stream.ReadUInt32();
                    string f2File = FileUtil.ReadPascalString(stream);
                    string f2Name = FileUtil.ReadPascalString(stream);
                    Console.WriteLine($"Warning: stub handler for type number: {tvt}: 0x{typedValue.Value:X} Function2 start:");
                    Console.WriteLine(string.Format("\tname = '{0}'", f2Name));
                    Console.WriteLine(string.Format("\tfile = '{0}'", f2File));
                    Console.WriteLine(string.Format("\tline = 0x{0} (decimal {0})", f2Line));
                    Console.WriteLine(string.Format("\tfp = 0x{0:X4}", f2FP));
                    Console.WriteLine(string.Format("\tsize = 0x{0:X8} (decimal {0})", f2Size));
                    Console.WriteLine(string.Format("\tretreg = 0x{0:X4}", f2RetReg));
                    Console.WriteLine(string.Format("\tmask = 0x{0:X8}", f2Mask));
                    Console.WriteLine(string.Format("\tmaskoffs = 0x{0:X8}", f2MaskOffset));
                    Console.WriteLine(string.Format("\tfmask = 0x{0:X8}", f2FMask));
                    Console.WriteLine(string.Format("\tfmaskoffs = 0x{0:X8}", f2FMaskOffset));
                    break;
                case 30:    // 0x1E
                    string mangledName1 = FileUtil.ReadPascalString(stream);
                    string mangledName2 = FileUtil.ReadPascalString(stream);
                    Console.WriteLine($"Warning: stub handler for type number: {tvt}: 0x{typedValue.Value:X} Mangled name: '{mangledName1}' is '{mangledName2}'");
                    break;
                default:
                    Console.WriteLine(string.Format("Error: Unhandled type number: {0} - data will be ignored, data after this point may be corrupted", tvt));
                    break;
            }
            return config.Tracker;
        }

        private void DumpType12(BinaryReader stream, int offset, SymDumpProcessingConfiguration config)
        {
            Function f = new Function(stream, (uint)offset, _funcTypes, config);
            f.SLDSourceFileName = config.Tracker.CurrentSLDFile;
            f.ObjectFileNames.Add(config.Tracker.CurrentObjectFileData.FileName);
            config.Tracker.CurrentObjectFileData.UpdateAddressData(f.Address);
            if (config.Tracker.CurrentOverlayID == -1)
            {
                Functions.Add(f);
            }
            else
            {
                config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Functions.Add(f);
            }
        }

        protected static Dictionary<string, EnumDef> ProcessEnum(EnumDef e, Dictionary<string, EnumDef> existingEnums,
            SymDumpProcessingConfiguration config)
        {
            EnumDef already;
            string name = e.Name;
            if (existingEnums.TryGetValue(name, out already))
            {
                string comparisonDesc = string.Format("Existing definition has hashcode 0x{0:X8} (dec: {0}), object files '{1}', SLD source file '{2}', and duplicate has hashcode 0x{3:X8} (dec: {3}), object files '{4}', SLD source file '{5}'.",
                    e.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(e.ObjectFileNames), e.SLDSourceFileName, 
                    already.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(already.ObjectFileNames), already.SLDSourceFileName);
                if (e.Equals(already))
                {
                    if (!config.IgnoreDuplicateDefinitions)
                    {
                        Console.WriteLine(string.Format("Info: duplicate definition of enum '{0}'. {1}", name, comparisonDesc));
                    }
                    return existingEnums;
                }
                else
                {
                    string originalName = name;
                    int iterationNum = 1;
                    while (existingEnums.TryGetValue(name, out already))
                    {
                        comparisonDesc = string.Format("Existing definition has hashcode 0x{0:X8} (dec: {0}), object files '{1}', SLD source file '{2}', and duplicate has hashcode 0x{3:X8} (dec: {3}), object files '{4}', SLD source file '{5}'.",
                            e.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(e.ObjectFileNames), e.SLDSourceFileName, 
                            already.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(already.ObjectFileNames), already.SLDSourceFileName);
                        if (e.Equals(already))
                        {
                            if (!config.IgnoreDuplicateDefinitions)
                            {
                                Console.WriteLine(string.Format("Info: duplicate definition of enum '{0}'. {1}", name, comparisonDesc));
                            }
                            return existingEnums;
                        }
                        name = String.Format("{0}_alt_definition_{1:D4}", originalName, iterationNum);
                        if (config.RenameForCompatibility)
                        {
                            name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(name, false);
                        }
                        iterationNum++;
                    }
                    Console.WriteLine($"WARNING: Non-uniform definitions of enum {originalName}. The redefined version has been renamed to {name}. {comparisonDesc}");
                    e.Name = name;
                }
            }

            if (config.RenameForCompatibility)
            {
                e.Name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(e.Name, false);
            }
            existingEnums.Add(e.Name, e);
            return existingEnums;
        }

        private void ReadEnum(BinaryReader reader, string name, SymDumpProcessingConfiguration config)
        {
            var e = new EnumDef(reader, name, config);
            e.SLDSourceFileName = config.Tracker.CurrentSLDFile;
            e.ObjectFileNames.Add(config.Tracker.CurrentObjectFileData.FileName);

            if (config.Tracker.CurrentOverlayID == -1)
            {
                _enums = ProcessEnum(e, _enums, config);
            }
            else
            {
                config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Enums = ProcessEnum(e, config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Enums, config);
            }
        }

        protected static Dictionary<string, UnionDef> ProcessUnion(UnionDef e, Dictionary<string, UnionDef> existingUnions, SymDumpProcessingConfiguration config)
        {
            //if (config.AutoRenameFakes && (e.IsFake))
            //{
            //    string lastObjectFileName = "";
            //    if (config.Tracker.ObjectFileDataEntries.Count > 0)
            //    {
            //        lastObjectFileName = config.Tracker.ObjectFileDataEntries[config.Tracker.ObjectFileDataEntries.Count - 1].FileName;
            //    }
            //    string lastSLDSourceFileName = "";
            //    if (config.Tracker.SLDEntries.Count > 0)
            //    {
            //        lastSLDSourceFileName = config.Tracker.SLDEntries[config.Tracker.SLDEntries.Count - 1].SLDFile;
            //    }
            //    e.Name = BenLincoln.DebugSymbols.Utilities.GetNewFakeName(e.Name, e.ObjectFileNames, config.Tracker.CurrentSLDFile, lastObjectFileName, lastSLDSourceFileName);
            //}
            //if (e.OriginalWasFake)
            //{
            //    config.FakeUnions.Add(e.Name, e);
            //}

            string unionName = e.Name;

            UnionDef already;
            if (existingUnions.TryGetValue(unionName, out already))
            {
                string comparisonDesc = string.Format("Existing definition has hashcode 0x{0:X8} (dec: {0}), object files '{1}', SLD source file '{2}', and duplicate has hashcode 0x{3:X8} (dec: {3}), object files '{4}', SLD source file '{5}'.", 
                    e.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(e.ObjectFileNames), e.SLDSourceFileName, already.GetHashCode(),
                    BenLincoln.DebugSymbols.Utilities.GetListAsString(already.ObjectFileNames), already.SLDSourceFileName);
                if (e.Equals(already))
                {
                    if (!config.IgnoreDuplicateDefinitions)
                    {
                        Console.WriteLine(string.Format("Info: duplicate definition of union '{0}'. {1}", unionName, comparisonDesc));
                    }
                    foreach (string objectFile in e.ObjectFileNames)
                    {
                        if (!already.ObjectFileNames.Contains(objectFile))
                        {
                            existingUnions[unionName].ObjectFileNames.Add(objectFile);
                            OutputDebugMessage(string.Format("Added object file '{0}' to existing union '{1}'", objectFile, unionName), config);
                        }
                    }
                    //foreach (string sourceFile in e.SourceFileNames)
                    //{
                    //    if (!already.ObjectFileNames.Contains(objectFile))
                    //    {
                    //        existingStructs[structName].ObjectFileNames.Add(objectFile);
                    //        OutputDebugMessage(string.Format("Added object file '{0}' to existing struct '{1}'", objectFile, structName), config);
                    //    }
                    //}
                    return existingUnions;
                }

                Console.WriteLine(string.Format("Warning: redefinition of union '{0}' with identical name. {1}", unionName, comparisonDesc));

                if (e.IsFake)
                {
                    // generate new "fake fake" name
                    var n = 1;

                    while (existingUnions.ContainsKey(String.Format("{0}_alt_definition_{1:D4}", unionName, n)))
                    {
                        n++;
                    }
                    string originalName = unionName;
                    unionName = String.Format("{0}_alt_definition_{1:D4}", unionName, n);
                    e.Name = unionName;

                    Console.WriteLine(string.Format("\tWarning: multiple 'fake' unions named '{0}'. The redefined version has been given the name '{1}'. Manual correction of the code related to this struct and/or its counterpart is likely to be required. {2}", originalName, unionName, comparisonDesc));
                }
                else
                {
                    string originalName = unionName;
                    int iterationNum = 1;
                    while (existingUnions.TryGetValue(unionName, out already))
                    {
                        comparisonDesc = string.Format("Existing definition has hashcode 0x{0:X8} (dec: {0}), object files '{1}', SLD source file '{2}', and duplicate has hashcode 0x{3:X8} (dec: {3}), object files '{4}', SLD source file '{5}'.",
                            e.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(e.ObjectFileNames), e.SLDSourceFileName, 
                            already.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(already.ObjectFileNames), already.SLDSourceFileName);
                        if (e.Equals(already))
                        {
                            if (!config.IgnoreDuplicateDefinitions)
                            {
                                Console.WriteLine(string.Format("\tInfo: duplicate definition of union '{0}'. {1}", unionName, comparisonDesc));
                            }
                            return existingUnions;
                        }
                        unionName = String.Format("{0}_alt_definition_{1:D4}", originalName, iterationNum);
                        if (config.RenameForCompatibility)
                        {
                            unionName = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(unionName, false);
                        }
                        iterationNum++;
                    }
                    Console.WriteLine($"\tWarning: Non-uniform definitions of union {originalName}. The redefined version has been given the key {unionName}. {comparisonDesc}");
                    e.Name = unionName;
                }
            }
            if (config.RenameForCompatibility)
            {
                e.Name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(e.Name, false);
            }

            for (int i = 0; i < e.Members.Count; i++)
            {
                e.Members[i].ParentName = e.Name;
                if (config.AutoRenameFakes && (e.Members[i].IsFake()))
                {
                    string lastObjectFileName = "";
                    if (config.Tracker.ObjectFileDataEntries.Count > 0)
                    {
                        lastObjectFileName = config.Tracker.ObjectFileDataEntries[config.Tracker.ObjectFileDataEntries.Count - 1].FileName;
                    }
                    string lastSLDSourceFileName = "";
                    if (config.Tracker.SLDEntries.Count > 0)
                    {
                        lastSLDSourceFileName = config.Tracker.SLDEntries[config.Tracker.SLDEntries.Count - 1].SLDFile;
                    }
                    e.Members[i].TypeInfo.Tag = BenLincoln.DebugSymbols.Utilities.GetNewFakeName("ProcessUnion", e.Members[i].TypeInfo.Tag, e.ObjectFileNames, config.Tracker.CurrentSLDFile, lastObjectFileName, lastSLDSourceFileName);
                }
            }

            existingUnions.Add(e.Name, e);

            return existingUnions;
        }

        private Dictionary<string, UnionDef> ReadUnion(Dictionary<string, UnionDef> unionDict, BinaryReader reader, 
            string name, SymDumpProcessingConfiguration config)
        {
            
            var e = new UnionDef(reader, name, config);
            e.SLDSourceFileName = config.Tracker.CurrentSLDFile;
            e.ObjectFileNames.Add(config.Tracker.CurrentObjectFileData.FileName);

            //unionDict = ProcessUnion(e, unionDict, config);
            config.AllUnions.Add(e);
            return unionDict;
        }

        protected static Dictionary<string, StructDef> ProcessStruct(StructDef e, Dictionary<string, StructDef> existingStructs,
            SymDumpProcessingConfiguration config)
        {
            //string unmodifiedName = e.Name;
            //if (config.AutoRenameFakes && (e.IsFake))
            //{
            //    string lastObjectFileName = "";
            //    if (config.Tracker.ObjectFileDataEntries.Count > 0)
            //    {
            //        lastObjectFileName = config.Tracker.ObjectFileDataEntries[config.Tracker.ObjectFileDataEntries.Count - 1].FileName;
            //    }
            //    string lastSLDSourceFileName = "";
            //    if (config.Tracker.SLDEntries.Count > 0)
            //    {
            //        lastSLDSourceFileName = config.Tracker.SLDEntries[config.Tracker.SLDEntries.Count - 1].SLDFile;
            //    }
            //    e.Name = BenLincoln.DebugSymbols.Utilities.GetNewFakeName(e.Name, e.ObjectFileNames, config.Tracker.CurrentSLDFile, lastObjectFileName, lastSLDSourceFileName);
            //}
            //string structName = e.Name;
            //if (e.OriginalWasFake)
            //{
            //    config.FakeStructs.Add(e.Name, e);
            //}

            string structName = e.Name;

            StructDef already;
            if (existingStructs.TryGetValue(structName, out already))
            {
                string comparisonDesc = string.Format("Existing definition has hashcode 0x{0:X8} (dec: {0}), object files '{1}', SLD source file '{2}', and duplicate has hashcode 0x{3:X8} (dec: {3}), object files '{4}', SLD source file '{5}'.",
                    e.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(e.ObjectFileNames), e.SLDSourceFileName, already.GetHashCode(), 
                    BenLincoln.DebugSymbols.Utilities.GetListAsString(already.ObjectFileNames), already.SLDSourceFileName);
                if (e.Equals(already))
                {
                    if (!config.IgnoreDuplicateDefinitions)
                    {
                        Console.WriteLine(string.Format("Info: duplicate definition of struct '{0}'. {1}", 
                            structName, comparisonDesc));
                    }
                    foreach (string objectFile in e.ObjectFileNames)
                    {
                        if (!already.ObjectFileNames.Contains(objectFile))
                        {
                            existingStructs[structName].ObjectFileNames.Add(objectFile);
                            OutputDebugMessage(string.Format("Added object file '{0}' to existing struct '{1}'", objectFile, structName), config);
                        }
                    }
                    //foreach (string sourceFile in e.SourceFileNames)
                    //{
                    //    if (!already.ObjectFileNames.Contains(objectFile))
                    //    {
                    //        existingStructs[structName].ObjectFileNames.Add(objectFile);
                    //        OutputDebugMessage(string.Format("Added object file '{0}' to existing struct '{1}'", objectFile, structName), config);
                    //    }
                    //}
                    return existingStructs;
                }
                Console.WriteLine(string.Format("Warning: redefinition of struct '{0}' with identical name.  {1}", 
                    structName, comparisonDesc));

                if (e.IsFake)
                {
                    // generate new "fake fake" name
                    var n = 1;
                    while (existingStructs.ContainsKey(String.Format("{0}_alt_definition_{1:D4}", structName, n)))
                    {
                        n++;
                    }
                    string originalName = structName;
                    structName = String.Format("{0}_alt_definition_{1:D4}", structName, n);
                    e.Name = structName;

                    Console.WriteLine(string.Format("\tWarning: multiple 'fake' structs named '{0}'. The redefined version has been given the name '{1}'. Manual correction of the code related to this struct and/or its counterpart is likely to be required. {2}", originalName, structName, comparisonDesc));
                }
                else
                {
                    string originalName = structName;
                    int iterationNum = 1;
                    while (existingStructs.TryGetValue(structName, out already))
                    {
                        comparisonDesc = string.Format("Existing definition has hashcode 0x{0:X8} (dec: {0}), object files '{1}', SLD source file '{2}', and duplicate has hashcode 0x{3:X8} (dec: {3}), object files '{4}', SLD source file '{5}'.",
                            e.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(e.ObjectFileNames), e.SLDSourceFileName, 
                            already.GetHashCode(), BenLincoln.DebugSymbols.Utilities.GetListAsString(already.ObjectFileNames), 
                            already.SLDSourceFileName);
                        if (e.Equals(already))
                        {
                            if (!config.IgnoreDuplicateDefinitions)
                            {
                                Console.WriteLine(string.Format("\tInfo: duplicate definition of struct '{0}'. {1}", structName, comparisonDesc));
                            }
                            return existingStructs;
                        }
                        structName = String.Format("{0}_alt_definition_{1:D4}", originalName, iterationNum);
                        if (config.RenameForCompatibility)
                        {
                            structName = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(structName, false);
                        }
                        iterationNum++;
                    }
                    e.Name = structName;
                    Console.WriteLine($"\tWarning: Non-uniform definitions of struct '{originalName}'. The redefined version has been given the name '{structName}'. {comparisonDesc}");
                }
            }
            if (config.RenameForCompatibility)
            {
                e.Name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(e.Name, false);
            }

            for (int i = 0; i < e.Members.Count; i++)
            {
                e.Members[i].ParentName = e.Name;
                if (config.AutoRenameFakes && (e.Members[i].IsFake()))
                {
                    string lastObjectFileName = "";
                    if (config.Tracker.ObjectFileDataEntries.Count > 0)
                    {
                        lastObjectFileName = config.Tracker.ObjectFileDataEntries[config.Tracker.ObjectFileDataEntries.Count - 1].FileName;
                    }
                    string lastSLDSourceFileName = "";
                    if (config.Tracker.SLDEntries.Count > 0)
                    {
                        lastSLDSourceFileName = config.Tracker.SLDEntries[config.Tracker.SLDEntries.Count - 1].SLDFile;
                    }
                    e.Members[i].TypeInfo.Tag = BenLincoln.DebugSymbols.Utilities.GetNewFakeName("ProcessStruct", e.Members[i].TypeInfo.Tag, e.ObjectFileNames, 
                        config.Tracker.CurrentSLDFile, lastObjectFileName, lastSLDSourceFileName);
                }
            }

            existingStructs.Add(structName, e);

            return existingStructs;
        }

        private Dictionary<string, StructDef> ReadStruct(Dictionary<string, StructDef> structDict, BinaryReader reader, 
            string name, SymDumpProcessingConfiguration config)
        {
            var e = new StructDef(reader, name, config);
            e.SLDSourceFileName = config.Tracker.CurrentSLDFile;
            e.ObjectFileNames.Add(config.Tracker.CurrentObjectFileData.FileName);
            //structDict = ProcessStruct(e, structDict, config);
            config.AllStructs.Add(e);
            return structDict;
        }

        protected static Dictionary<string, TypeInfo> ProcessTypedef(Dictionary<string, TypeInfo> existingTypedefs, string name, TypeInfo typeInfo,
            SymDumpProcessingConfiguration config)
        {
            if (config.AutoRenameFakes)
            {
                if (typeInfo.IsFake)
                {
                    string lastObjectFileName = "";
                    if (config.Tracker.ObjectFileDataEntries.Count > 0)
                    {
                        lastObjectFileName = config.Tracker.ObjectFileDataEntries[config.Tracker.ObjectFileDataEntries.Count - 1].FileName;
                    }
                    string lastSLDSourceFileName = "";
                    if (config.Tracker.SLDEntries.Count > 0)
                    {
                        lastSLDSourceFileName = config.Tracker.SLDEntries[config.Tracker.SLDEntries.Count - 1].SLDFile;
                    }
                    typeInfo.Tag = BenLincoln.DebugSymbols.Utilities.GetNewFakeName("ProcessTypedef", typeInfo.Tag, typeInfo.ObjectFileNames, config.Tracker.CurrentSLDFile, lastObjectFileName, lastSLDSourceFileName);
                }
            }

            TypeInfo already;
            if (existingTypedefs.TryGetValue(name, out already))
            {
                if (typeInfo.Equals(already))
                {
                    if (!config.IgnoreDuplicateDefinitions)
                    {
                        Console.WriteLine(string.Format("Info: duplicate definition of typedef '{0}'. Existing definition has hashcode 0x{1:X8} (dec: {1}), and duplicate has hashcode 0x{2:X8} (dec: {2}).", name, typeInfo.GetHashCode(), already.GetHashCode()));
                    }
                    return existingTypedefs;
                }
                else
                {
                    //throw new Exception($"Non-uniform definitions of typedef for {name}");
                    string originalName = name;
                    int iterationNum = 1;
                    while (existingTypedefs.TryGetValue(name, out already))
                    {
                        if (typeInfo.Equals(already))
                        {
                            if (!config.IgnoreDuplicateDefinitions)
                            {
                                Console.WriteLine(string.Format("Info: duplicate definition of typedef '{0}'. Existing definition has hashcode 0x{1:X8} (dec: {1}), and duplicate has hashcode 0x{2:X8} (dec: {2}).", name, typeInfo.GetHashCode(), already.GetHashCode()));
                            }
                            return existingTypedefs;
                        }
                        name = String.Format("{0}_alt_definition_{1:D4}", originalName, iterationNum);
                        if (config.RenameForCompatibility)
                        {
                            name = BenLincoln.DebugSymbols.SymDump.CElement.GetCompatibleName(name, false);
                        }
                        iterationNum++;
                    }
                    Console.WriteLine($"WARNING: Non-uniform definitions of typedef for {originalName}. The redefined version has been renamed to {name}.");
                    // not sure if I need to rename the typedef somehow or not
                }
            }
            existingTypedefs.Add(name, typeInfo);

            return existingTypedefs;
        }

        private void AddTypedef(string name, TypeInfo typeInfo, SymDumpProcessingConfiguration config)
        {
            if (config.Tracker.CurrentOverlayID == -1)
            {
                _typedefs = ProcessTypedef(_typedefs, name, typeInfo, config);
            }
            else
            {
                config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Typedefs = ProcessTypedef(config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Typedefs, 
                    name, typeInfo, config);
            }
        }

        private void DumpType20(BinaryReader stream, int offset, SymDumpProcessingConfiguration config)
        {
            long streamPosition = stream.BaseStream.Position;
            var ti = stream.ReadTypeInfo(false, config);
            var name = stream.ReadPascalString();
            ti.SLDSourceFileName = config.Tracker.CurrentSLDFile;
            ti.ObjectFileNames.Add(config.Tracker.CurrentObjectFileData.FileName);

            if (ti.ClassType == ClassType.Enum && ti.TypeDef.BaseType == BaseType.EnumDef)
            {
                OutputDebugMessage(string.Format("Entry type 20, subtype 'Enum': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                ReadEnum(stream, name, config);
                return;
            }

            if (ti.ClassType == ClassType.FileName)
            {
                OutputDebugMessage(string.Format("Entry type 20, subtype 'FileName': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                config.Tracker.HandleObjectFileFileName(name);
                return;
            }
            if (ti.ClassType == ClassType.Struct && ti.TypeDef.BaseType == BaseType.StructDef)
            {
                OutputDebugMessage(string.Format("Entry type 20, subtype 'Struct': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                
                if (config.Tracker.CurrentOverlayID == -1)
                {
                    _structs = ReadStruct(_structs, stream, name, config);
                }
                else
                {
                    config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Structs = ReadStruct(config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Structs, stream, name, config);
                }
            }
            else if (ti.ClassType == ClassType.Union && ti.TypeDef.BaseType == BaseType.UnionDef)
            {
                OutputDebugMessage(string.Format("Entry type 20, subtype 'Union': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                if (config.Tracker.CurrentOverlayID == -1)
                {
                    _unions = ReadUnion(_unions, stream, name, config);
                }
                else
                {
                    config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Unions = ReadUnion(config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Unions, 
                        stream, name, config);
                }
            }
            else if (ti.ClassType == ClassType.Typedef)
            {
                OutputDebugMessage(string.Format("Entry type 20, subtype 'Typedef': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                AddTypedef(name, ti, config);
            }
            else if (ti.ClassType == ClassType.External)
            {
                OutputDebugMessage(string.Format("Entry type 20, subtype 'Extern': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                if (ti.TypeDef.IsFunctionReturnType)
                {
                    OutputDebugMessage("\tThis is a function return type, so adding to the list of function types", config);
                    _funcTypes[name] = ti.AsCode("").Trim();
                }
                else
                {
                    OutputDebugMessage("\tThis is not a function return type, so adding to the list of externs", config);
                    //_externs.Add($"extern {ti.AsCode(name)}; // offset 0x{offset:X}");
                    BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x = new BenLincoln.DebugSymbols.SymDump.CExternalDeclaration();
                    x.name = name;
                    x.declaration = $"extern {ti.AsCode(name)}; // offset 0x{offset:X}";
                    x.class_type = ti.ClassType.ToString();
                    x.offset_in_memory = (uint)offset;
                    x.size = ti.Size;                    
                    x.extern_type = "extern";
                    if (config.Tracker.CurrentOverlayID == -1)
                    {
                        _externs.Add(x);
                    }
                    else
                    {
                        config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Externs.Add(x);
                    }
                }
            }
            else if (ti.ClassType == ClassType.Static)
            {
                OutputDebugMessage(string.Format("Entry type 20, subtype 'Static': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                if (ti.TypeDef.IsFunctionReturnType)
                {
                    OutputDebugMessage("\tThis is a function return type, so adding to the list of function types", config);
                    _funcTypes[name] = ti.AsCode("").Trim();
                }
                else
                {
                    OutputDebugMessage("\tThis is not a function return type, so adding to the list of externs", config);
                    //_externs.Add($"static {ti.AsCode(name)}; // offset 0x{offset:X}");
                    BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x = new BenLincoln.DebugSymbols.SymDump.CExternalDeclaration();
                    x.name = name;
                    x.declaration = $"static {ti.AsCode(name)}; // offset 0x{offset:X}";
                    x.class_type = ti.ClassType.ToString();
                    x.offset_in_memory = (uint)offset;
                    x.size = ti.Size;
                    x.extern_type = "static";
                    x.source_file = config.Tracker.CurrentSLDFile;
                    x.object_files.Add(config.Tracker.CurrentObjectFileData.FileName);
                    config.Tracker.CurrentObjectFileData.UpdateAddressData((uint)x.offset_in_memory);
                    if (config.Tracker.CurrentOverlayID == -1)
                    {
                        _externs.Add(x);
                    }
                    else
                    {
                        config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Externs.Add(x);
                    }
                }
            }
            else
            {
                Console.WriteLine(string.Format("Error: Unhandled type 20 subtype {0} @0x{1:X8} - data will be ignored, data after this point may be corrupted", ti.ClassType, streamPosition));
            }
        }

        private void DumpType22(BinaryReader stream, int offset, SymDumpProcessingConfiguration config)
        {
            long streamPosition = stream.BaseStream.Position;
            var ti = stream.ReadTypeInfo(true, config);
            var name = stream.ReadPascalString();
            ti.SLDSourceFileName = config.Tracker.CurrentSLDFile;
            ti.ObjectFileNames.Add(config.Tracker.CurrentObjectFileData.FileName);

            if (ti.ClassType == ClassType.Enum && ti.TypeDef.BaseType == BaseType.EnumDef)
            {
                OutputDebugMessage(string.Format("Entry type 22, subtype 'Enum': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                ReadEnum(stream, name, config);
            }
            else if (ti.ClassType == ClassType.Typedef)
            {
                OutputDebugMessage(string.Format("Entry type 22, subtype 'Typedef': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                AddTypedef(name, ti, config);
            }
            else if (ti.ClassType == ClassType.External)
            {
                OutputDebugMessage(string.Format("Entry type 22, subtype 'Extern': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                if (ti.TypeDef.IsFunctionReturnType)
                {
                    OutputDebugMessage("\tThis is a function return type, so adding to the list of function types", config);
                    _funcTypes[name] = ti.AsCode("").Trim();
                }
                else
                {
                    OutputDebugMessage("\tThis is not a function return type, so adding to the list of externs", config);
                    //_externs.Add($"extern {ti.AsCode(name)}; // offset 0x{offset:X}");
                    BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x = new BenLincoln.DebugSymbols.SymDump.CExternalDeclaration();
                    x.name = name;
                    x.declaration = $"extern {ti.AsCode(name)}; // offset 0x{offset:X}";
                    x.class_type = ti.ClassType.ToString();
                    x.offset_in_memory = (uint)offset;
                    x.size = ti.Size;
                    x.extern_type = "extern";
                    x.source_file = config.Tracker.CurrentSLDFile;
                    x.object_files.Add(config.Tracker.CurrentObjectFileData.FileName);
                    config.Tracker.CurrentObjectFileData.UpdateAddressData((uint)x.offset_in_memory);
                    if (config.Tracker.CurrentOverlayID == -1)
                    {
                        _externs.Add(x);
                    }
                    else
                    {
                        config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Externs.Add(x);
                    }
                }
            }
            else if (ti.ClassType == ClassType.Static)
            {
                OutputDebugMessage(string.Format("Entry type 22, subtype 'Static': address 0x{0:X8}, name '{1}' @0x{2:X8}", offset, name, streamPosition), config);
                if (ti.TypeDef.IsFunctionReturnType)
                {
                    OutputDebugMessage("\tThis is a function return type, so adding to the list of function types", config);
                    _funcTypes[name] = ti.AsCode("").Trim();
                }
                else
                {
                    OutputDebugMessage("\tThis is not a function return type, so adding to the list of externs", config);
                    //_externs.Add($"static {ti.AsCode(name)}; // offset 0x{offset:X}");
                    BenLincoln.DebugSymbols.SymDump.CExternalDeclaration x = new BenLincoln.DebugSymbols.SymDump.CExternalDeclaration();
                    x.name = name;
                    x.declaration = $"static {ti.AsCode(name)}; // offset 0x{offset:X}";
                    x.class_type = ti.ClassType.ToString();
                    x.offset_in_memory = (uint)offset;
                    x.size = ti.Size;
                    x.extern_type = "static";
                    x.source_file = config.Tracker.CurrentSLDFile;
                    x.object_files.Add(config.Tracker.CurrentObjectFileData.FileName);
                    config.Tracker.CurrentObjectFileData.UpdateAddressData((uint)x.offset_in_memory);
                    if (config.Tracker.CurrentOverlayID == -1)
                    {
                        _externs.Add(x);
                    }
                    else
                    {
                        config.Tracker.Overlays[config.Tracker.CurrentOverlayID].Externs.Add(x);
                    }
                }
            }
            else
            {
                Console.WriteLine(string.Format("Error: Unhandled type 22 subtype {0} @0x{1:X8} - data will be ignored, data after this point may be corrupted", ti.ClassType, streamPosition));
            }
        }

        public Function FindFunction(uint addr)
        {
            // return Functions.FirstOrDefault(f => f.Address == addr);
            foreach (Function f in Functions)
            {
                if (f.Address == addr)
                {
                    return f;
                }
            }
            foreach (int overlayID in _overlays.Keys)
            {
                PsyQOverlay ol = _overlays[overlayID];
                foreach (Function f in ol.Functions)
                {
                    if (f.Address == addr)
                    {
                        return f;
                    }
                }
            }
            throw new Exception(string.Format("Error: FindFunction: Could not find a function with address 0x{0:X8}", addr));
        }
    }
}
