﻿using System;
using System.IO;

namespace symdump.symfile
{
    public class TypedValue : IEquatable<TypedValue>
    {
        public readonly byte Type;
        public readonly int Value;
        public readonly int SYMFileSequenceNumber;
        public string ObjectFileName;
        protected SymDumpProcessingConfiguration ProcessingConfig;

        public TypedValue(BinaryReader fs, SymDumpProcessingConfiguration config)
        {
            Value = fs.ReadInt32();
            Type = fs.ReadByte();
            SYMFileSequenceNumber = config.Issuer.GetNextNumber();
            ProcessingConfig = config;
        }

        public bool IsLabel => (Type & 0x80) == 0;

        public bool Equals(TypedValue other)
        {
            if (ReferenceEquals(null, other)) return false;
            if (ReferenceEquals(this, other)) return true;
            if (Type != other.Type)
            {
                SymFile.OutputDebugMessage(string.Format("TypedValue objects don't have matching Type ('{0}' vs '{1}')", Type.ToString(), other.Type.ToString()), ProcessingConfig);
                return false;
            }
            if (Value != other.Value)
            {
                SymFile.OutputDebugMessage(string.Format("TypedValue objects don't have matching Value ('{0}' vs '{1}')", Value.ToString(), other.Value.ToString()), ProcessingConfig);
                return false;
            }
            return true;
        }

        public override string ToString()
        {
            return $"value={Value} type={Type} isLabel={IsLabel}";
        }

        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj)) return false;
            if (ReferenceEquals(this, obj)) return true;
            if (obj.GetType() != GetType()) return false;
            return Equals((TypedValue) obj);
        }

        public override int GetHashCode()
        {
            unchecked
            {
                return (Type.GetHashCode() * 397) ^ Value;
            }
        }
    }
}
