﻿using System;
using System.IO;

namespace symdump.symfile.util
{
    public static class FileUtil
    {
        public static void Skip(this BinaryReader s, int n)
        {
            s.BaseStream.Seek(n, SeekOrigin.Current);
        }

        public static string ReadPascalString(this BinaryReader fs)
        {
            var length = fs.ReadByte();
            var result = "";
            if (length > 0)
            {
                int position = length;
                while (position-- > 0)
                {
                    try
                    {
                        //result += fs.ReadChar();
                        // ^^^ doesn't work on high-value ASCII characters, like Japanese?
                        // pretty sure there are no PSX games that use UTF encoding
                        byte currentByte = fs.ReadByte();
                        result += (char)currentByte;
                    }
                    catch (Exception e)
                    {
                        Console.WriteLine(string.Format("Exception thrown when reading position {0} of Pascal string with length {1}: {2}", position, length, e.Message));
                    }
                }
            }
            return result;
        }

        public static TypeDef ReadTypeDef(this BinaryReader s, BenLincoln.DebugSymbols.SequenceIssuer issuer)
        {
            return new TypeDef(s, issuer);
        }

        public static TypeInfo ReadTypeInfo(this BinaryReader s, bool withDimensions, SymDumpProcessingConfiguration config)
        {
            return new TypeInfo(s, withDimensions, config);
        }

        public static ClassType ReadClassType(this BinaryReader s)
        {
            return (ClassType) s.ReadUInt16();
        }

        public static bool SkipSld(this BinaryReader reader, TypedValue typedValue)
        {
            switch (typedValue.Type & 0x7f)
            {
                case 0:
                    return true;
                case 2:
                    reader.Skip(1);
                    return true;
                case 4:
                    reader.Skip(2);
                    return true;
                case 6:
                    reader.Skip(4);
                    return true;
                case 8:
                    reader.Skip(4);
                    reader.Skip(reader.ReadByte());
                    return true;
                case 10:
                    return true;
            }

            return false;
        }
    }
}
